////////////////////////////////////////////////////////////////////////////////
// Lists and Stacks
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit HIT_Lists;

//==============================================================================
{.$DEFINE DEBUG}
//------------------------------------------------------------------------------
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Classes, Integers, GF2;

//==============================================================================
// The following classes are very simplistic. They are only bases that you can
// improve depending on what you need. If you do, the lists should never contain
// NIL pointers.
//==============================================================================

type
  //===== Abstract ancestor class
  TAbstractHITList = class(THITObject)
  private
    fX: TList;
    function GetCount: Integer;
  public
    constructor Create;
    destructor Destroy; override;
    //-----
    procedure Flush; virtual; abstract;
    procedure Sort(Compare: TListSortCompare);
    //-----
    property Count: Integer read GetCount;
  end;


  /// INTEGERS ///////////////////////////////////////////////////////////////

  //===== Abstract ancestor class (Integers)
  TAbstractIntList = class(TAbstractHITList)
  public
    procedure Flush; override;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
  end;

  //===== TIntStack class
  TIntStack = class(TAbstractIntList)
  public
    function Pop(A: TInteger): Boolean;
    procedure Push(A: TInteger);
  end;

  //===== TIntList class
  TIntList = class(TAbstractIntList)
  private
    function GetPtr(I: Integer): TInteger;
  public
    procedure Append(A: TInteger);
    procedure Delete(I: Integer);
    function IndexOf(A: TInteger): Integer;
    procedure Insert(I: Integer; A: TInteger);
    //-----
    property X[I: Integer]: TInteger read GetPtr; default;
  end;


  /// POLYNOMIALS OVER GF(2) /////////////////////////////////////////////////

  //===== Abstract ancestor class
  TAbstractGF2List = class(TAbstractHITList)
  public
    procedure Flush; override;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
  end;

  //===== TGF2Stack class
  TGF2Stack = class(TAbstractGF2List)
  public
    function Pop(P: TGF2Poly): Boolean;
    procedure Push(P: TGF2Poly);
  end;

  //===== TGF2List class
  TGF2List = class(TAbstractGF2List)
  private
    function GetPtr(I: Integer): TGF2Poly;
  public
    procedure Append(P: TGF2Poly);
    procedure Delete(I: Integer);
    function IndexOf(P: TGF2Poly): Integer;
    procedure Insert(I: Integer; P: TGF2Poly);
    //-----
    property X[I: Integer]: TGF2Poly read GetPtr; default;
  end;


/// MISCELLANEOUS ////////////////////////////////////////////////////////////

function GF2ListCmp(P, Q: TGF2Poly): Integer;
function IListCmp(A, B: TInteger): Integer;
function IListCmpAbs(A, B: TInteger): Integer;

implementation
//==============================================================================
{$C-,G-,O+,Q-,R-}
//==============================================================================
uses SysUtils;

////////////////////////////////////////////////////////////////////////////////
// Comparison functions
////////////////////////////////////////////////////////////////////////////////

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// The following functions are necessary in order to call the TList.Sort
// method. When using the call "Sort(@ICmp)", it would seem that DELPHI is not
// able to select the right overloaded ICmp function. ICmp is declared in the
// Integers unit. Same for GF2Cmp
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//==============================================================================
//==============================================================================
function GF2ListCmp(P, Q: TGF2Poly): Integer;
begin
  Result := GF2Cmp(P,Q);
end;

//==============================================================================
//==============================================================================
function IListCmp(A, B: TInteger): Integer;
begin
  Result := ICmp(A,B);
end;

//==============================================================================
//==============================================================================
function IListCmpAbs(A, B: TInteger): Integer;
begin
  Result := ICmpAbs(A,B);
end;

////////////////////////////////////////////////////////////////////////////////
// TYPES, CONSTS
////////////////////////////////////////////////////////////////////////////////

//----- Used by the LoadFromFile and the SaveToFile methods
//      !!! DO NOT MODIFY
type
  THITListHeaderFile = array[0..16] of AnsiChar;
const
  INTLIST_HEADER_FILE : THITListHeaderFile = 'HIT Integer List ';
  GF2LIST_HEADER_FILE : THITListHeaderFile = 'HIT GF2Poly List ';

////////////////////////////////////////////////////////////////////////////////
// TAbstractHITList class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
//==============================================================================
constructor TAbstractHITList.Create;
begin
  fX := TList.Create;
end;

//==============================================================================
//==============================================================================
destructor TAbstractHITList.Destroy;
begin
  //----- Free the HIT objects
  Flush;
  //----- Free the list
  fX.Free;
end;

//==============================================================================
// Return the count of objects that are in the list
//==============================================================================
function TAbstractHITList.GetCount: Integer;
begin
  Result := fX.Count;
end;

//==============================================================================
// Sort the list
// If Self is a list INTEGERS, do not use the ICmp and ICmpAbs functions of
// Integers because of overload problem. Used the IListCmp and IListCmpAbs
// functions:
// "L.Sort(@IListCmp)" or "L.Sort(@IListCmpAbs)"
//==============================================================================
procedure TAbstractHITList.Sort(Compare: TListSortCompare);
begin
  fX.Sort(Compare);
end;

////////////////////////////////////////////////////////////////////////////////
// TAbstractIntList class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Flush the list
//==============================================================================
procedure TAbstractIntList.Flush;
  var i : Integer;
begin
  with fX do
  begin
    for i := 0 to Count-1 do TInteger(fX[i]).Free;
    Clear;
  end;
end;

//==============================================================================
// Load a list from a file
// If fX is not empty, it is flushed.
//==============================================================================
procedure TAbstractIntList.LoadFromFile(const FileName: string);
  const
    FLAGS = fmOpenRead or fmShareDenyWrite;
  var
    F      : TFileStream;
    Header : THITListHeaderFile;
begin
  F := TFileStream.Create(ExpandFileName(FileName),FLAGS);
  try
    //----- Read the header of the file
    F.ReadBuffer(Header,SizeOf(Header));
    if Header <> INTLIST_HEADER_FILE then
      HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
    //----- Load the contents of the list
    Flush;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do
      begin
        Flush;
        HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
      end;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
// Append to the list the contents of a stream
//==============================================================================
procedure TAbstractIntList.LoadFromStream(Stream: TStream);
  var
    Delta : Longint; //! Must be signed and 4-byte sized
    i, j  : Integer;
begin
  //----- Number of elements
  Stream.ReadBuffer(Delta,SizeOf(Delta));
  //----- Adjust the size of the "Items" table of fX
  fX.Capacity := fX.Count + Delta;
  //----- Load
  for i := 0 to Pred(Delta) do
  begin
    //----- Create a number
    j := fX.Add(TInteger.Create);
    //----- Load the number
    TInteger(fX[j]).LoadFromStream(Stream);
  end;
end;

//==============================================================================
// Save the list to a file
//==============================================================================
procedure TAbstractIntList.SaveToFile(const FileName: string);
  const FLAGS = fmCreate or fmShareExclusive;
  var F: TFileStream;
begin
  F := TFileStream.Create(ExpandFileName(FileName),FLAGS);
  try
    //----- Write the header of the file
    F.WriteBuffer(INTLIST_HEADER_FILE,SizeOf(INTLIST_HEADER_FILE));
    //----- Save the contents of the list
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Save the list to a stream
//==============================================================================
procedure TAbstractIntList.SaveToStream(Stream: TStream);
  var i : Longint; //! Must be signed and 4-byte sized
begin
  //----- Save fX.Count
  i := fX.Count;
  Stream.WriteBuffer(i,SizeOf(i));
  //----- Save values
  for i := 0 to fX.Count-1 do TInteger(fX[i]).SaveToStream(Stream);
end;

////////////////////////////////////////////////////////////////////////////////
// TIntStack class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Pop the integer that is on the top of the stack
// Return TRUE if an integer is popped, otherwise return FALSE
// "Pop(nil)" allows to drop a number
//==============================================================================
function TIntStack.Pop(A: TInteger): Boolean;
begin
  with fX do
  if Count = 0 then Result := false
  else
  begin
    if Assigned(A) then ISwp(A,fX[0]); //"ISwp" is faster than "ISet"
    TInteger(fX[0]).Free;
    Delete(0);
    Capacity := Count;
    Result := true;
  end;
end;

//==============================================================================
// Push a copy of an integer (on the top of the stack)
// Do nothing if A = nil
//==============================================================================
procedure TIntStack.Push(A: TInteger);
begin
  if Assigned(A) then fX.Insert(0,TInteger.Create(A));
end;

////////////////////////////////////////////////////////////////////////////////
// TIntList class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Append a copy of a number to the end of the list
// Do nothing if A = nil
//==============================================================================
procedure TIntList.Append(A: TInteger);
begin
  if Assigned(A) then fX.Add(TInteger.Create(A));
end;

//==============================================================================
// Delete the integer A at the position I in the list
//==============================================================================
procedure TIntList.Delete(I: Integer);
begin
  if (I < 0) or (I >= fX.Count) then HITRaiseIndexError(ClassName + '.Delete');

  with fX do
  begin
    TInteger(fX[I]).Free;
    Delete(I);
    Capacity := Count;
  end;
end;

//==============================================================================
// Access method used by the X property
//==============================================================================
function TIntList.GetPtr(I: Integer): TInteger;
begin
  if (I < 0) or (I >= fX.Count) then HITRaiseIndexError(ClassName + '.GetPtr');

  Result := fX[I];
end;

//==============================================================================
// Search the position of the integer A in the list
// Return -1 if A = nil or if A is not in the list
//==============================================================================
function TIntList.IndexOf(A: TInteger): Integer;
begin
  if Assigned(A) then
  with fX do
  begin
    Result := 0;
    while (Result < Count) and (ICmp(TInteger(fX[Result]),A) <> 0) do
      Inc(Result);
    if Result = Count then Result := -1;
  end
  else Result := -1;
end;

//==============================================================================
// Insert a copy of an integer in the list at the position I
// Do nothing if A = nil
//==============================================================================
procedure TIntList.Insert(I: Integer; A: TInteger);
begin
  if I < 0 then HITRaiseIndexError(ClassName + '.Insert');

  if Assigned(A) then
  with fX do
  begin
    //----- To append the number at the end of the list
    if I > Count then I := Count;
    Insert(I,TInteger.Create(A));
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// TAbstractGF2List class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Flush the list
//==============================================================================
procedure TAbstractGF2List.Flush;
  var i : Integer;
begin
  with fX do
  begin
    for i := 0 to Count-1 do TGF2Poly(fX[i]).Free;
    Clear;
  end;
end;

//==============================================================================
// Load a list from a file
// If fX is not empty, it is flushed.
//==============================================================================
procedure TAbstractGF2List.LoadFromFile(const FileName: string);
  const
    FLAGS = fmOpenRead or fmShareDenyWrite;
  var
    F      : TFileStream;
    Header : THITListHeaderFile;
begin
  F := TFileStream.Create(ExpandFileName(FileName),FLAGS);
  try
    //----- Read the header of the file
    F.ReadBuffer(Header,SizeOf(Header));
    if Header <> GF2LIST_HEADER_FILE then
      HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
    //----- Load the contents of the list
    Flush;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do
      begin
        Flush;
        HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
      end;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
// Append to the list the contents of a stream
//==============================================================================
procedure TAbstractGF2List.LoadFromStream(Stream: TStream);
  var
    Delta : Longint; //! Must be signed and 4-byte sized
    i, j  : Integer;
begin
  //----- Number of elements
  Stream.ReadBuffer(Delta,SizeOf(Delta));
  //----- Adjust the size of the "Items" table of fX
  fX.Capacity := fX.Count + Delta;
  //----- Load
  for i := 0 to Pred(Delta) do
  begin
    //----- Create a number
    j := fX.Add(TGF2Poly.Create);
    //----- Load the polynomial
    TGF2Poly(fX[j]).LoadFromStream(Stream);
  end;
end;

//==============================================================================
// Save the list to a file
//==============================================================================
procedure TAbstractGF2List.SaveToFile(const FileName: string);
  const FLAGS = fmCreate or fmShareExclusive;
  var F: TFileStream;
begin
  F := TFileStream.Create(ExpandFileName(FileName),FLAGS);
  try
    //----- Write the header of the file
    F.WriteBuffer(GF2LIST_HEADER_FILE,SizeOf(GF2LIST_HEADER_FILE));
    //----- Save the contents of the list
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Save the list to a stream
//==============================================================================
procedure TAbstractGF2List.SaveToStream(Stream: TStream);
  var i : Longint; //! Must be signed and 4-byte sized
begin
  //----- Save fX.Count
  i := fX.Count;
  Stream.WriteBuffer(i,SizeOf(i));
  //----- Save values
  for i := 0 to fX.Count-1 do TGF2Poly(fX[i]).SaveToStream(Stream);
end;

////////////////////////////////////////////////////////////////////////////////
// TGF2Stack class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Pop the polynomial that is on the top of the stack
// Return TRUE if a polynomial is popped, otherwise return FALSE
// "Pop(nil)" allows to drop a polynomial
//==============================================================================
function TGF2Stack.Pop(P: TGF2Poly): Boolean;
begin
  with fX do
  if Count = 0 then Result := false
  else
  begin
    if Assigned(P) then GF2Swp(P,fX[0]); //"GF2Swp" is faster than "GF2Set"
    TGF2Poly(fX[0]).Free;
    Delete(0);
    Capacity := Count;
    Result := true;
  end;
end;

//==============================================================================
// Push a copy of a polynomial (on the top of the stack)
// Do nothing if P = nil
//==============================================================================
procedure TGF2Stack.Push(P: TGF2Poly);
begin
  if Assigned(P) then fX.Insert(0,TGF2Poly.Create(P));
end;

////////////////////////////////////////////////////////////////////////////////
// TGF2List class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Append a copy of a polynomial to the end of the list
// Do nothing if P = nil
//==============================================================================
procedure TGF2List.Append(P: TGF2Poly);
begin
  if Assigned(P) then fX.Add(TGF2Poly.Create(P));
end;

//==============================================================================
// Delete the polynomial at the position I in the list
//==============================================================================
procedure TGF2List.Delete(I: Integer);
begin
  if (I < 0) or (I >= fX.Count) then HITRaiseIndexError(ClassName + '.Delete');

  with fX do
  begin
    TGF2Poly(fX[I]).Free;
    Delete(I);
    Capacity := Count;
  end;
end;

//==============================================================================
// Access method used by the X property
//==============================================================================
function TGF2List.GetPtr(I: Integer): TGF2Poly;
begin
  if (I < 0) or (I >= fX.Count) then HITRaiseIndexError(ClassName + '.GetPtr');

  Result := fX[I];
end;

//==============================================================================
// Search the position of the polynomial P in the list
// Return -1 if P = nil or if P is not in the list
//==============================================================================
function TGF2List.IndexOf(P: TGF2Poly): Integer;
begin
  if Assigned(P) then
  with fX do
  begin
    Result := 0;
    while (Result < Count) and not GF2Equ(TGF2Poly(fX[Result]),P) do
      Inc(Result);
    if Result = Count then Result := -1;
  end
  else Result := -1;
end;

//==============================================================================
// Insert a copy of a polynomial in the list at the position I
// Do nothing if P = nil
//==============================================================================
procedure TGF2List.Insert(I: Integer; P: TGF2Poly);
begin
  if I < 0 then HITRaiseIndexError(ClassName + '.Insert');

  if Assigned(P) then
  with fX do
  begin
    //----- To append the polynomial at the end of the list
    if I > Count then I := Count;
    Insert(I,TGF2Poly.Create(P));
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

