////////////////////////////////////////////////////////////////////////////////
// Cryptographically Secure Random Generator
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit HIT_Rand;

//==============================================================================
{.$DEFINE DEBUG}
//------------------------------------------------------------------------------
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers, GF2;

type
  //----- Ancestor of HIT Cryptographically Secure Random Generators
  TAbstractHITCSRG = class(THITObject)
  private
    fSeed: TInteger;
    fCount: Longword;
    //-----
    procedure RandSeed;
    procedure Update;
  public
    constructor Create;
    destructor  Destroy; override;
  end;

  //----- INTEGERS
  TIntCSRG = class(TAbstractHITCSRG)
  public
    procedure NewSeed(S: TInteger = nil);
    procedure Rnd(A: TInteger; BitCount: Longint; HiBit: Boolean = true);
  end;

  //----- GF2
  TGF2CSRG = class(TAbstractHITCSRG)
  public
    procedure NewSeed(S: TGF2Poly = nil);
    procedure Rnd(P: TGF2Poly; Deg: Longint; HiDeg: Boolean = true);
  end;

implementation
//==============================================================================
uses HIT_Hash, Windows;

////////////////////////////////////////////////////////////////////////////////
// RESOURCE STRINGS
////////////////////////////////////////////////////////////////////////////////

resourcestring
  sINTMsg101 = '("BitCount" < 0) or ("BitCount" > TInteger.MaxBitSize)';
  sGF2Msg610 = '("Deg" < 0) or ("Deg" > TGF2Poly.MaxDegree)';


////////////////////////////////////////////////////////////////////////////////
// TAbstractHITCSRG CLASS
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
//==============================================================================
constructor TAbstractHITCSRG.Create;
begin
  fSeed := TInteger.Create;
  RandSeed;
end;

//==============================================================================
//==============================================================================
destructor TAbstractHITCSRG.Destroy;
begin
  fSeed.Free;
end;

//------------------------------------------------------------------------------
// Modify the internal state of the generator
// Make use of the polynomial [512, 510, 507, 504, 0] over GF(2)
// $a4800000 = 2^31 + 2^29 + 2^26 + 2^23
//------------------------------------------------------------------------------
procedure TAbstractHITCSRG.Update;
begin
  //----- LFSR
  if IShr(fSeed) then fSeed.Long[15] := fSeed.Long[15] xor $a4800000;
end;

//------------------------------------------------------------------------------
// Seed the generator with a true (not reproducible) random seed
//------------------------------------------------------------------------------
procedure TAbstractHITCSRG.RandSeed;
  type TL4 = array[0..3] of Longword;
  var
    M : TMemoryStatus;
    T : TSystemTime;
    X : array[0..15] of Longword;
begin
  try
    repeat
      GetSystemTime(T);
      M.dwLength := SizeOf(M);
      GlobalMemoryStatus(M);
      X[15] := TL4(T)[3];
      X[14] := M.dwMemoryLoad;
      X[13] := TL4(T)[2];
      X[12] := M.dwTotalPhys;
      X[11] := TL4(T)[1];
      X[10] := M.dwAvailPhys;
      X[ 9] := TL4(T)[0];
      X[ 8] := M.dwTotalPageFile;
      X[ 7] := GetCurrentProcessID;
      X[ 6] := M.dwAvailPageFile;
      X[ 5] := GetCurrentThreadID;
      X[ 4] := M.dwTotalVirtual;
      X[ 3] := fCount;
      X[ 2] := M.dwAvailVirtual;
      X[ 1] := GetTickCount;
      X[ 0] := Longword(GetMessageTime);
      fSeed.SetDigits(X,SizeOf(X));
    until not fSeed.Is0;
  finally
    FillChar(T,SizeOf(T),0);
    FillChar(M,SizeOf(M),0);
    FillChar(X,SizeOf(X),0);
  end;
  fCount := 0;
end;

////////////////////////////////////////////////////////////////////////////////
// TIntCSRG CLASS
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Set the generator with a new seed
// If S is nil, the generator is randomly seeded
//==============================================================================
procedure TIntCSRG.NewSeed(S: TInteger = nil);
begin
  if not Assigned(S) then RandSeed
  else
  begin
    ISet(fSeed,S);
    repeat
      IH160(fSeed,512,Ord(S.Neg));
    until not fSeed.Is0;
    fCount := 0;
  end;
end;

//==============================================================================
// <- A, BitCount-bit random integer
//==============================================================================
procedure TIntCSRG.Rnd(A: TInteger; BitCount: Longint; HiBit: Boolean = true);
begin
  with A do
  if BitCount = 0 then Set0
  else
  if (BitCount > 0) and (BitCount <= MaxBitSize) then
  begin
    ISet(A,fSeed);
    Inc(fCount);
    IH128(A,BitCount,fCount);
    //----- Set the leading bit (if need be)
    if HiBit then Bit[BitCount-1] := true;
    //----- Modify fSeed
    Update;
  end
  else HITRaiseInvalidArg(ClassName + '.Rnd',LoadResString(@sINTMsg101));
end;

////////////////////////////////////////////////////////////////////////////////
// TGF2RG CLASS
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Set the generator with a new seed
// If S is nil, the generator is randomly seeded
//==============================================================================
procedure TGF2CSRG.NewSeed(S: TGF2Poly = nil);
begin
  if not Assigned(S) then RandSeed
  else
  begin
    GF2PolyToInt(S,fSeed);
    repeat
      IH160(fSeed,512,0);
    until not fSeed.Is0;
    fCount := 0;
  end;
end;

//==============================================================================
// <- P, random polynomial over GF(2) of degree Deg
//==============================================================================
procedure TGF2CSRG.Rnd(P: TGF2Poly; Deg: Longint; HiDeg: Boolean = true);
begin
  with P do
  if (Deg >= 0) and (Deg <= MaxDegree) then
  begin
    GF2IntToPoly(fSeed,P);
    Inc(fCount);
    GF2H128(P,Deg,fCount);
    //----- Set the higher degree (if need be)
    if HiDeg then Coef[Deg] := true;
    //----- Modify fSeed
    Update;
  end
  else HITRaiseInvalidArg(ClassName + '.Rnd',LoadResString(@sGF2Msg610));
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

