unit ElList;

interface

uses Classes, SysUtils;

type
  TElListDeleteEvent = procedure(Sender : TObject; Item : Pointer) of object;

  TElList = class
  private
    FList: PPointerList;
    FCount: Integer;
    FCapacity: Integer;
    FOnDelete : TElListDeleteEvent;
  protected
    function Get(Index: Integer): Pointer;
    procedure Grow; virtual;
    procedure Put(Index: Integer; Item: Pointer);
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetCount(NewCount: Integer);
    procedure TriggerDeleteEvent(Item : Pointer); virtual;
  public
    destructor Destroy; override;
    function Add(Item: Pointer): Integer;
    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    function Expand: TElList;
    function First: Pointer;
    function IndexOf(Item: Pointer): Integer;
    procedure Insert(Index: Integer; Item: Pointer);
    function Last: Pointer;
    procedure Move(CurIndex, NewIndex: Integer);
    function Remove(Item: Pointer): Integer;
    procedure Pack;
    procedure Sort(Compare: TListSortCompare);
    property Capacity: Integer read FCapacity write SetCapacity;
    property Count: Integer read FCount write SetCount;
    property Items[Index: Integer]: Pointer read Get write Put; default;
    property List: PPointerList read FList;
    property OnDelete : TElListDeleteEvent read FOnDelete write FOnDelete;
  end;

implementation

destructor TElList.Destroy;
begin
  Clear;
  inherited;
end;

function TElList.Add(Item: Pointer): Integer;
begin
  Result := FCount;
  if Result = FCapacity then Grow;
  FList^[Result] := Item;
  Inc(FCount);
end;

procedure TElList.Clear;
var I : integer;

begin
  for i:=0 to Count -1 do TriggerDeleteEvent(Get(i));
  SetCount(0);
  SetCapacity(0);
end;

procedure TElList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then raise EListError.Create('List index out of bounds.');
  TriggerDeleteEvent(Get(Index));
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(Pointer));
  if FCount < (FCapacity div 2) then SetCapacity (FCapacity div 2);
end;

procedure TElList.Exchange(Index1, Index2: Integer);
var
  Item: Pointer;
begin
  if (Index1 < 0) or (Index1 >= FCount) then raise EListError.Create('List index out of bounds.');
  if (Index2 < 0) or (Index2 >= FCount) then raise EListError.Create('List index out of bounds.');
  Item := FList^[Index1];
  FList^[Index1] := FList^[Index2];
  FList^[Index2] := Item;
end;

function TElList.Expand: TElList;
begin
  if FCount = FCapacity then Grow;
  Result := Self;
end;

function TElList.First: Pointer;
begin
  Result := Get(0);
end;

function TElList.Get(Index: Integer): Pointer;
begin
  if (Index < 0) or (Index >= FCount) then raise EListError.Create('List index out of bounds.');
  Result := FList^[Index];
end;

procedure TElList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TElList.IndexOf(Item: Pointer): Integer;
begin
  Result := 0;
  while (Result < FCount) and (FList^[Result] <> Item) do Inc(Result);
  if Result = FCount then Result := -1;
end;

procedure TElList.Insert(Index: Integer; Item: Pointer);
begin
  if (Index < 0) or (Index > FCount) then raise EListError.Create('List index out of bounds.');
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(Pointer));
  FList^[Index] := Item;
  Inc(FCount);
end;

function TElList.Last: Pointer;
begin
  Result := Get(FCount - 1);
end;

procedure TElList.Move(CurIndex, NewIndex: Integer);
var
  Item: Pointer;
begin
  if CurIndex <> NewIndex then
  begin
    if (NewIndex < 0) or (NewIndex >= FCount) then raise EListError.Create('List index out of bounds.');
    Item := Get(CurIndex);
    Delete(CurIndex);
    Insert(NewIndex, Item);
  end;
end;

procedure TElList.Put(Index: Integer; Item: Pointer);
begin
  if (Index < 0) or (Index >= FCount) then raise EListError.Create('List index out of bounds.');
  FList^[Index] := Item;
end;

function TElList.Remove(Item: Pointer): Integer;
begin
  Result := IndexOf(Item);
  if Result <> -1 then Delete(Result);
end;

procedure TElList.Pack;
var
  I: Integer;
begin
  for I := FCount - 1 downto 0 do if Items[I] = nil then Delete(I);
end;

procedure TElList.SetCapacity(NewCapacity: Integer);
begin
  if (NewCapacity < FCount) or (NewCapacity > MaxListSize) then
     raise EListError.Create('List index out of bounds.');
  if NewCapacity <> FCapacity then
  begin
    ReallocMem(FList, NewCapacity * SizeOf(Pointer));
    FCapacity := NewCapacity;
  end;
end;

procedure TElList.SetCount(NewCount: Integer);
begin
  if (NewCount < 0) or (NewCount > MaxListSize) then
    raise EListError.Create('List index out of bounds.');
  if NewCount > FCapacity then SetCapacity(NewCount);
  if NewCount > FCount then
    FillChar(FList^[FCount], (NewCount - FCount) * SizeOf(Pointer), 0);
  FCount := NewCount;
end;

procedure QuickSort(SortList: PPointerList; L, R: Integer;
  SCompare: TListSortCompare);
var
  I, J: Integer;
  P, T: Pointer;
begin
  repeat
    I := L;
    J := R;
    P := SortList^[(L + R) shr 1];
    repeat
      while SCompare(SortList^[I], P) < 0 do Inc(I);
      while SCompare(SortList^[J], P) > 0 do Dec(J);
      if I <= J then
      begin
        T := SortList^[I];
        SortList^[I] := SortList^[J];
        SortList^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(SortList, L, J, SCompare);
    L := I;
  until I >= R;
end;

procedure TElList.Sort(Compare: TListSortCompare);
begin
  if (FList <> nil) and (Count > 0) then
    QuickSort(FList, 0, Count - 1, Compare);
end;

procedure TElList.TriggerDeleteEvent(Item : Pointer);
{ Triggers the OnDelete event. This is a virtual method (descendants of this component can override it). }
{ CDK: Call as needed to trigger event. }
begin
  if (assigned(FOnDelete)) then
    FOnDelete(Self, Item );
end;  { TriggerDeleteEvent }

end.
