////////////////////////////////////////////////////////////////////////////////
// ElGamal Signature Scheme
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
//------------------------------------------------------------------------------
// Concerning this signature scheme, an interesting article written by
// D. Bleichenbacher, "Generating ElGamal signatures without knowing the secret
// key" is downloadable from
// ftp://ftp.inf.ethz.ch/pub/publications/papers/ti/isc/ElGamal.ps
////////////////////////////////////////////////////////////////////////////////
unit ElGamal;

//==============================================================================
{.$DEFINE DEBUG}
//------------------------------------------------------------------------------
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers;

function  EGCheck(R,S,H,P,G,Y: TInteger): Boolean;
procedure EGKeys(P,G,X,Y: TInteger; BSize: Longint);
procedure EGSign(R,S,H,P,G,X: TInteger);

implementation
//==============================================================================
uses Integers_Primes, HIT_Rand;

//------------------------------------------------------------------------------
// Return TRUE iff the P pointers are distinct
//------------------------------------------------------------------------------
function DistinctPointers(const P: array of Pointer): Boolean;
  var i, j : Longint;
begin
  Result := true;
  for i := 0 to High(P)-1 do
    for j := i+1 to High(P) do
      if P[i] = P[j] then
      begin
        Result := false;
        Exit;
      end;
end;

//==============================================================================
// Compute the public an secret keys for the ElGamal scheme
// -> BSize, binary size of P
// <- P, prime (Safe prime, P = 2Q + 1 with Q prime)
// <- G, generator
// <- X, part of secret key
// <- Y, part of public key
//------------------------------------------------------------------------------
// Public key: P,G,Y
// Secret key: P,G,X
//==============================================================================
procedure EGKeys(P,G,X,Y: TInteger; BSize: Longint);
  var
    rg    : TIntCSRG; //Secure random generator (See HIT_Rand.pas)
    OldCL : Boolean;
begin
  //----- Check
  if not DistinctPointers([P,G,X,Y]) then
    HITRaiseDuplicatePtr('EGKeys','P,G,X,Y');

  rg := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;
    //----- Random generator
    rg := TIntCSRG.Create;

    //----- Safe prime P (the creation may be long enough)
    rg.Rnd(P,BSize);
    IMakeSafePrime(P,[2,3,5,7,11]);

    //----- Generator G
    repeat
      rg.Rnd(G,BSize+1);
      IMod(G,P);
    until IJacobi(G,P) = -1;

    //----- Generate X and Y
    rg.Rnd(X,BSize+1);
    IMod(X,P);
    ISet(Y,G);
    IPowMod(Y,X,P);
  finally
    //----- Free the generator
    rg.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

//==============================================================================
// ElGamal signature of the message digest H
// <- R,S, signature of H
// -> H, message digest
// -> P,G,X secret key of the signer
//==============================================================================
procedure EGSign(R,S,H,P,G,X: TInteger);
  var
    rg         : TIntCSRG; //Secure random generator (See HIT_Rand.pas)
    K, PMinus1 : TInteger;
    BSize      : Longint;
    OldCL      : Boolean;
begin
  //----- Check
  if not DistinctPointers([R,S,H,P,G,X]) then
    HITRaiseDuplicatePtr('EGSign','R,S,H,P,G,X');

  rg := nil;
  K := nil;
  PMinus1 := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;
    //----- Create a secure random generator
    rg := TIntCSRG.Create;

    //----- PMinus1 := P-1
    PMinus1 := TInteger.Create(P);
    IDec(PMinus1);

    //----- Exception if (P-1) <= H
    if ICmp(PMinus1,H) < 1 then
      HITRaiseInvalidArg('EGSign','H should be less than P-1');

    //----- K random, K > Log2(P) and relatively prime to P-1
    K := TInteger.Create;
    BSize := P.BitSize;
    repeat
      //----- K, random and secure
      rg.Rnd(K,BSize+1);
      K.Odd := true; //So that the remainder is odd
      IMod(K,PMinus1);
    until (ICmp(K,BSize) > 0) and IGCD1(K,PMinus1);
    //----- Signature R = G^K mod P
    ISet(R,G);
    IPowMod(R,K,P);
    //----- Signature S = ((H-R*X) * K^(-1)) mod (P-1)
    ISet(S,R);
    IMul(S,X);
    IMod(S,PMinus1);
    ISubr(S,H);
    IInvMod(K,PMinus1);
    IMul(S,K);
    //----- To always return a nonnegative S
    IMod(S,PMinus1);
  finally
    //----- Free the generator
    rg.Free;
    //----- Free the numbers
    K.Free;
    PMinus1.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

//==============================================================================
// Check an ElGamal signature
// -> R,S, signature of H
// -> H, message digest
// -> P,G,Y, public key of the signer
// <- Result = TRUE if the signature (R,S) is valid, otherwise FALSE
//==============================================================================
function EGCheck(R,S,H,P,G,Y: TInteger): Boolean;
  var
    A, B  : TInteger;
    OldCL : Boolean;
begin
  //----- Check
  if not DistinctPointers([R,S,H,P,G,Y]) then
    HITRaiseDuplicatePtr('EGSign','R,S,H,P,G,Y');

  Result := false;
  if R.Is0 or (ICmp(R,P) >= 0) then Exit; //Invalid signature
  A := nil;
  B := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;

    A := TInteger.Create(P);
    //----- Exception if (P-1) <= H
    IDec(A);
    if ICmp(A,H) < 1 then
      HITRaiseInvalidArg('EGCheck','H should be less than P-1');

    //----- A = Y^R * R^S mod P
    ISet(A,Y);
    IPowMod(A,R,P);
    B := TInteger.Create(R);
    IPowMod(B,S,P);
    IMul(A,B);
    IMod(A,P);
    //----- B = G^H mod P
    ISet(B,G);
    IPowMod(B,H,P);
    //----- If A = B then the signature is valid
    Result := IEqu(A,B);
  finally
    A.Free;
    B.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

