////////////////////////////////////////////////////////////////////////////////
// Elliptic Curves over GF(2^N)
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit EllipticCurvesGF2N;

//==============================================================================
{.$DEFINE DEBUG}
//------------------------------------------------------------------------------
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G-,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Classes, Integers, GF2;

//==============================================================================
// The elliptic curves over GF(2^N) used here are always
// y^2 + xy + x^3 + Ax^2 + B = 0
// Such a curve is non supersingular if and only if B <> 0.
// Since (0,0) cannot be a valid point, it is used as "point at infinity".
//==============================================================================

type
  //===== Point of a curve y^2 + xy + x^3 + Ax^2 + B = 0 over GF(2^N)
  //      Affine coordinates
  //      Point at infinity O = (0,0)
  TGF2NEPoint = class(THITObject)
  private
    fX: TGF2Poly;
    fY: TGF2Poly;
  public
    constructor Create; overload;
    constructor Create(P: TGF2NEPoint); overload;
    destructor  Destroy; override;
    //-----
    procedure Copy(P: TGF2NEPoint);
    function  Equ(P: TGF2NEPoint): Boolean;
    function  IsO: Boolean;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure SetO;
    procedure SetXY(X, Y: TGF2Poly);
    //-----
    property X: TGF2Poly read fX; //Don't use it to modify fX
    property Y: TGF2Poly read fY; //Don't use it to modify fY
  end;

  //===== Curve (Weierstrass form)
  //      y^2 + xy + x^3 + Ax^2 + B = 0 over GF(2^N)
  TGF2NECurveStatus = (csEmpty, csHasField, csHasCoefs);
  TGF2NECurve = class(THITObject)
  private
    fF: TGF2N;    //Field GF(2^N)
    fA: TGF2Poly; //Coef, polynomial over GF(2)
    fB: TGF2Poly; //Coef, polynomial over GF(2)
    fC: TGF2Poly; //Aux, equal to fB^(2^(fF.ModulusDegree-2))
    fStatus: TGF2NECurveStatus;
    //-----
    procedure ErrorNoCoefs(const Caller: string);
    procedure ErrorNoField(const Caller: string);
    procedure JAdd(X0,Y0,Z0, X1,Y1,Z1: TGF2Poly);
    procedure JDouble(X,Y,Z: TGF2Poly);
    procedure JToAC(JX,JY,JZ: TGF2Poly; P: TGF2NEpoint);
  public
    constructor Create; overload;
    constructor Create(C: TGF2NECurve); overload;
    destructor  Destroy; override;
    //-----
    procedure Add(P, Q: TGF2NEPoint);
    procedure Copy(C: TGF2NECurve);
    procedure Double(P: TGF2NEPoint);
    procedure GetCoefs(A, B: TGF2Poly);
    procedure GetField(F: TGF2N);
    procedure GetModulus(M: TGF2Poly);
    procedure GetOrder2Point(P: TGF2NEPoint);
    procedure GetRndPoint(P: TGF2NEPoint);
    function  Halves(H1, H2, P: TGF2NEPoint): Boolean;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure __LoadFromStream(Stream: TStream); virtual;
    procedure Mul(P: TGF2NEPoint; K: TInteger; Jacobian: Boolean = false);
    procedure Negate(P: TGF2NEPoint); overload;
    function  OnCurve(P: TGF2NEPoint): Boolean;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure SetCoefs(const A, B: array of Longint); overload;
    procedure SetCoefs(A, B: TGF2Poly); overload;
    procedure SetField(const Degrees: array of Longint; Check: Boolean); overload;
    procedure SetField(F: TGF2N); overload;
    procedure SetField(P: TGF2Poly; Check: Boolean); overload;
    procedure SetRndCoefs(BOnly: Boolean = false);
    procedure Sub(P, Q: TGF2NEPoint);
    //-----
    property  Status: TGF2NECurveStatus read fStatus;
  end;

implementation
//==============================================================================
uses SysUtils;

////////////////////////////////////////////////////////////////////////////////
// CONSTS, VARS, TYPES, ETC.
////////////////////////////////////////////////////////////////////////////////

const
  //----- File flags
  LOADFLAGS = fmOpenRead or fmShareDenyWrite;
  SAVEFLAGS = fmCreate or fmShareExclusive;

type
  TEPHeaderFile = array[0..24] of AnsiChar;
  TECHeaderFile = array[0..24] of AnsiChar;

const
  //----- Header of "point" file
  EPHeaderFile : TEPHeaderFile = 'HIT EC over GF2N - Point ';
  //----- Header of "curve" file
  ECHeaderFile : TECHeaderFile = 'HIT EC over GF2N - Curve ';


////////////////////////////////////////////////////////////////////////////////
// TGF2NEPoint CLASS
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.Copy(P: TGF2NEPoint);
begin
  GF2Set(fX,P.fX);
  GF2Set(fY,P.fY);
end;

//==============================================================================
//==============================================================================
constructor TGF2NEPoint.Create;
begin
  fX := TGF2Poly.Create;
  fY := TGF2Poly.Create;
end;

//==============================================================================
//==============================================================================
constructor TGF2NEPoint.Create(P: TGF2NEPoint);
begin
  fX := TGF2Poly.Create(P.fX);
  fY := TGF2Poly.Create(P.fY);
end;

//==============================================================================
//==============================================================================
destructor TGF2NEPoint.Destroy;
begin
  fX.Free;
  fY.Free;
end;

//==============================================================================
// Result := TRUE iff Self = P
//==============================================================================
function TGF2NEPoint.Equ(P: TGF2NEPoint): Boolean;
begin
  Result := GF2Equ(fX,P.fX) and GF2Equ(fY,P.fY);
end;

//==============================================================================
// Return TRUE iff the point is the "point at infinity"
//==============================================================================
function TGF2NEPoint.IsO: Boolean;
begin
  Result := fX.Is0 and fY.Is0;
end;

//==============================================================================
// Load Self from a file
//==============================================================================
procedure TGF2NEPoint.LoadFromFile(const FileName: string);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------------------------
  var
    F : TFileStream;
    H : TEPHeaderFile;
begin
  F := TFileStream.Create(FileName,LOADFLAGS);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> EPHeaderFile then Error;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.LoadFromStream(Stream: TStream);
begin
  fX.LoadFromStream(Stream);
  fY.LoadFromStream(Stream);
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,SAVEFLAGS);
  try
    F.WriteBuffer(EPHeaderFile,SizeOf(EPHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.SaveToStream(Stream: TStream);
begin
  fX.SaveToStream(Stream);
  fY.SaveToStream(Stream);
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.SetO;
begin
  fX.Set0;
  fY.Set0;
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.SetXY(X, Y: TGF2Poly);
begin
  GF2Set(fX,X);
  GF2Set(fY,Y);
end;


////////////////////////////////////////////////////////////////////////////////
// TGF2NECurve CLASS
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Add
// -> P, Q, points
// <- P = P + Q
//==============================================================================
procedure TGF2NECurve.Add(P, Q: TGF2NEPoint);
  var S, T : TGF2Poly;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Add');

  //----- If Q = O, there is nothing to do
  if Q.IsO then Exit;
  //----- If P = O then P := Q
  if P.IsO then
  begin
    P.Copy(Q);
    Exit;
  end;
  //----- PX = QX ?
  if GF2Equ(P.fX,Q.fX) then
  begin
    //----- If PY <> QY then return P = O
    if not GF2Equ(P.fY,Q.fY) then P.SetO
    else
    //----- If PX = 0 then return P = O (PX = 0 -> P is 2-order point)
    if P.fX.Is0 then P.SetO
    else Double(P);
    Exit;
  end;

  T := nil;
  S := TGF2Poly.Create(P.fX);
  try
    //----- S := (PY + QY)/(PX + QX)
    fF.Add(S,Q.fX);
    fF.Inv(S);
    T := TGF2Poly.Create(P.fY);
    fF.Add(T,Q.fY);
    fF.Mul(S,T);
    //----- T = S + S + QX + A
    GF2Set(T,S);
    fF.Sqr(T);
    fF.Add(T,S);
    fF.Add(T,Q.fX);
    fF.Add(T,fA);
    //----- PX' = PX + T
    fF.Add(P.fX,T);
    //----- PY' = PY + PX' + ST
    fF.Mul(T,S);
    fF.Add(P.fY,T);
    fF.Add(P.fY,P.fX);
  finally
    S.Free;
    T.Free;
  end;
end;

//==============================================================================
// Set Self equal to C
//==============================================================================
procedure TGF2NECurve.Copy(C: TGF2NECurve);
begin
  fF.Copy(C.fF);
  GF2Set(fA,C.fA);
  GF2Set(fB,C.fB);
  GF2Set(fC,C.fC);
  fStatus := C.fStatus;
end;

//==============================================================================
//==============================================================================
constructor TGF2NECurve.Create;
begin
  fF := TGF2N.Create;
  fA := TGF2Poly.Create;
  fB := TGF2Poly.Create;
  fC := TGF2Poly.Create;
end;

//==============================================================================
// Create Self equal to C
//==============================================================================
constructor TGF2NECurve.Create(C: TGF2NECurve);
begin
  fF := TGF2N.Create(C.fF);
  fA := TGF2Poly.Create(C.fA);
  fB := TGF2Poly.Create(C.fB);
  fC := TGF2Poly.Create(C.fC);
  fStatus := C.fStatus;
end;

//==============================================================================
//==============================================================================
destructor TGF2NECurve.Destroy;
begin
  fF.Free;
  fA.Free;
  fB.Free;
  fC.Free;
end;

//==============================================================================
// Double
// -> P, point
// <- P = P x 2
//==============================================================================
procedure TGF2NECurve.Double(P: TGF2NEPoint);
  var S, T : TGF2Poly;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Double');

  with P do
  begin
    //----- If PX = O, there is nothing to do
    if fX.Is0 then //Handle the case P = O
    begin
      fY.Set0; //Here, equivalent to P.SetO
      Exit;
    end;
    S := nil;
    T := nil;
    try
      S := TGF2Poly.Create(fX);
      T := TGF2Poly.Create(fX);
      //----- T := PX + PY/PX
      fF.Inv(T);
      fF.Mul(T,fY);
      fF.Add(T,fX);
      //----- PX' = T + T + A
      GF2Set(fX,T);
      fF.Sqr(fX);
      fF.Add(fX,T);
      fF.Add(fX,fA);
      //----- PY' = PY + PX' + (PX + PX')T
      fF.Add(fY,fX);
      fF.Add(S,fX);
      fF.Mul(S,T);
      fF.Add(fY,S);
    finally
      S.Free;
      T.Free;
    end;
  end;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure TGF2NECurve.ErrorNoCoefs(const Caller: string);
begin
  HITRaiseInvalidCall(ClassName + '.' + Caller,'Curve without coefs');
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure TGF2NECurve.ErrorNoField(const Caller: string);
begin
  HITRaiseInvalidCall(ClassName + '.' + Caller,'Curve without field');
end;

//==============================================================================
// If you need only B, for instance, use "GetCoefs(nil,X);"
//==============================================================================
procedure TGF2NECurve.GetCoefs(A, B: TGF2Poly);
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('GetCoefs');

  if Assigned(A) then GF2Set(A,fA);
  if Assigned(B) then GF2Set(B,fB);
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.GetField(F: TGF2N);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('GetField');

  F.Copy(fF);
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.GetModulus(M: TGF2Poly);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('GetModulus');

  fF.GetModulus(M);
end;

//==============================================================================
// Return the unique 2-torsion point of the curve
// <- P = (0,B^(1/2))
//==============================================================================
procedure TGF2NECurve.GetOrder2Point(P: TGF2NEPoint);
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('GetOrder2Point');

  with P do
  begin
    fX.Set0;
    GF2Set(fY,fB);
    fF.SqrRoot(fY);
  end;
end;

//==============================================================================
// Generate a random point
// <- P = (X,Y), random point different from (0,0)
//==============================================================================
procedure TGF2NECurve.GetRndPoint(P: TGF2NEPoint);
  var U : TGF2Poly;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('GetRndPoint');

  U := TGF2Poly.Create;
  try
    with P do
    repeat
      fF.Rnd(fX);
      if fX.Is0 then
      begin
        GetOrder2Point(P);
        Exit;
      end;
      //----- U = (PX + A)PX + B
      GF2Set(U,fX);
      fF.Add(U,fA);
      GF2Set(fY,fX);
      fF.Sqr(fY);
      fF.Mul(U,fY);
      fF.Add(U,fB);
      if U.Is0 then
      begin
        fY.Set0;
        Break;
      end;
    until fF.SolveQuad(fY,fX,U);
    //----- Toss up: P or -P ?
    fF.Rnd(U);
    if U.Parity then Negate(P);
  finally
    U.Free;
  end;
end;

//==============================================================================
// Compute H1 and H2 such that P = 2*H1 = 2*H2 (if such halves exist)
// Return TRUE iff the halves exist.
//------------------------------------------------------------------------------
// The used algorithm is trivial, so I suppose it is in the public domain. But,
// if ever it was not, and if it is not already patented, I claim it is my
// intellectual property and I put it in the public domain.
// The purpose of this note? Only to prevent any patent.
//==============================================================================
function TGF2NECurve.Halves(H1, H2, P: TGF2NEPoint): Boolean;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Halves');

  //----- Compute T = X + Y/X (X and Y are unknow, they are what we want)
  //      Solve f(T) = T + T + A + Px
  H2.fX.Set1;
  GF2Set(H2.fY,fA);
  fF.Add(H2.fY,P.fX);
  Result := fF.SolveQuad(H1.fX,H2.fX,H2.fY); //H1.fX is used for "T"
  if not Result then
  begin
    H1.SetO; //Cleaner
    H2.SetO; //"
    Exit;
  end;

  //----- The 2 roots T and T+1 lead to the 2 halves H1 and H2
  GF2Set(H2.fY,H1.fX);
  GF2Set(H1.fY,H1.fX);
  H1.fY[0] := not H1.fY[0];
  GF2Set(H2.fX,H1.fY);

  //----- Compute H1.X
  //      X := (T Px + Py)^(1/2)
  fF.Mul(H1.fX,P.fX);
  fF.Add(H1.fX,P.fY);
  fF.SqrRoot(H1.fX);
  //----- Compute H1.Y
  //      Y := X (X + T + 1)
  fF.Add(H1.fY,H1.fX);
  fF.Mul(H1.fY,H1.fX);

  //----- Compute H2.X
  //      X := ((T + 1) Px + Py)^(1/2)
  fF.Mul(H2.fX,P.fX);
  fF.Add(H2.fX,P.fY);
  fF.SqrRoot(H2.fX);
  //----- Compute H2.Y
  //      Y := X (X + T)
  fF.Add(H2.fY,H2.fX);
  fF.Mul(H2.fY,H2.fX);
end;

//------------------------------------------------------------------------------
// Adding in Jacobian coordinates
// -> (X0,Y0,Z0) and (X1,Y1,Z1), points of the curve
// <- (X0,Y0,Z0) = (X0,Y0,Z0) + (X1,Y1,Z1)
//------------------------------------------------------------------------------
// The IEEE P1363 Annex A calls "projective" what is called "Jacobian" here.
// The Jacobian coordinates are projective ones, but the word "projective" is
// ordinary given to an other kind of projective coordinates. In fact, in most
// articles or books you will read, "projective" will not be equivalent to
// "projective" in the IEEE P1363.
//------------------------------------------------------------------------------
procedure TGF2NECurve.JAdd(X0,Y0,Z0, X1,Y1,Z1: TGF2Poly);
  var T1, T2, T3, T4 : TGF2Poly;
begin
  T1 := nil;
  T2 := nil;
  T3 := nil;
  T4 := nil;
  try
    T1 := TGF2Poly.Create(X1);
    T2 := TGF2Poly.Create(Y1);
    T3 := TGF2Poly.Create;
    T4 := TGF2Poly.Create(X1);
    if not Z1.Is1 then
    begin
      GF2Set(T3,Z1);
      fF.Sqr(T3);
      fF.Mul(X0,T3);
      fF.Mul(T3,Z1);
      fF.Mul(Y0,T3);
    end;
    GF2Set(T3,Z0);
    fF.Sqr(T3);
    fF.Mul(T4,T3);
    fF.Add(X0,T4);
    fF.Mul(T3,Z0);
    fF.Mul(T3,Y1);
    fF.Add(Y0,T3);
    if X0.Is0 then
    begin
      if not Y0.Is0 then
      begin
        X0.Set1;
        Y0.Set1;
      end;
      Z0.Set0;
      Exit;
    end;
    fF.Mul(T1,Y0);
    fF.Mul(Z0,X0);
    fF.Mul(T2,Z0);
    fF.Add(T1,T2);
    GF2Set(T2,Z0);
    fF.Sqr(T2);
    GF2Set(T3,T1);
    fF.Mul(T3,T2);
    if not Z1.Is1 then fF.Mul(Z0,Z1);
    GF2Set(T1,Y0);
    fF.Add(T1,Z0);
    fF.Mul(Y0,T1);
    GF2Set(T2,X0);
    fF.Sqr(T2);
    fF.Mul(X0,T2);
    if not fA.Is0 then
    begin
      GF2Set(T2,Z0);
      fF.Sqr(T2);
      fF.Mul(T2,fA);
      fF.Add(X0,T2);
    end;
    fF.Add(X0,Y0);
    fF.Mul(T1,X0);
    GF2Swp(Y0,T1); //Faster than "GF2Set"
    fF.Add(Y0,T3);
  finally
    T1.Free;
    T2.Free;
    T3.Free;
    T4.Free;
  end;
end;

//------------------------------------------------------------------------------
// Doubling in Jacobian coordinates
// -> (X,Y,Z) point of the curve
// <- (X,Y,Z) = 2(X,Y,Z)
//------------------------------------------------------------------------------
procedure TGF2NECurve.JDouble(X,Y,Z: TGF2Poly);
  var T : TGF2Poly;
begin
  //----- Handle special case
  if X.Is0 or Z.Is0 then
  begin
    X.Set1;
    Y.Set1;
    Z.Set0;
    Exit;
  end;

  T := TGF2Poly.Create(fC);
  try
    fF.Mul(Y,Z);
    fF.Sqr(Z);
    fF.Mul(T,Z);
    fF.Mul(Z,X);
    fF.Add(Y,Z);
    fF.Add(T,X);
    fF.Sqr(T);
    fF.Sqr(T);
    fF.Sqr(X);
    fF.Add(Y,X);
    fF.Mul(Y,T);
    fF.Sqr(X);
    fF.Mul(X,Z);
    fF.Add(Y,X);
    GF2Swp(X,T); //Faster than "GF2Set"
  finally
    T.Free;
  end;
end;

//------------------------------------------------------------------------------
// Conversion: Jacobian coordinates to affine coordinates
//------------------------------------------------------------------------------
procedure TGF2NECurve.JToAC(JX,JY,JZ: TGF2Poly; P: TGF2NEpoint);
  var U, V : TGF2Poly;
begin
  if JZ.Is0 then
  begin
    P.fX.Set0;
    P.fY.Set0;
  end
  else
  begin
    V := nil;
    U := TGF2Poly.Create(JZ);
    try
      fF.Inv(U);
      V := TGF2Poly.Create(U);
      fF.Sqr(U);
      GF2Set(P.fX,JX);
      fF.Mul(P.fX,U);  //AX := JX / JZ^2
      fF.Mul(U,V);
      GF2Set(P.fY,JY);
      fF.Mul(P.fY,U);  //AY := JY / JZ^3
    finally
      U.Free;
      V.Free;
    end;
  end;
end;

//==============================================================================
// Load Self from a file
//==============================================================================
procedure TGF2NECurve.LoadFromFile(const FileName: string);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------------------------
  var
    F : TFileStream;
    H : TECHeaderFile;
begin
  F := TFileStream.Create(FileName,LOADFLAGS);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> ECHeaderFile then Error;
    try
      LoadFromStream(F);
      // Use the following instruction instead of the previous one
      // in order to load files saved with previous version
      //__LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.LoadFromStream(Stream: TStream);
begin
  Stream.ReadBuffer(fStatus,SizeOf(fStatus));
  fF.LoadFromStream(Stream);
  fA.LoadFromStream(Stream);
  fB.LoadFromStream(Stream);
  fC.LoadFromStream(Stream);
end;

//==============================================================================
// Load from stream of a previous HIT release (that had no fC field)
//==============================================================================
procedure TGF2NECurve.__LoadFromStream(Stream: TStream);
begin
  Stream.ReadBuffer(fStatus,SizeOf(fStatus));
  fF.LoadFromStream(Stream);
  fA.LoadFromStream(Stream);
  fB.LoadFromStream(Stream);
  //-----
  GF2Set(fC,fB);
  fF.RepeatedSqr(fC,fF.ModulusDegree-2);
end;

//==============================================================================
// Mul
// -> P, point
// -> K, integer
// -> Jacobian, if TRUE use Jacobian coordinates, if FALSE use affine ones
// <- P = P x K
//==============================================================================
procedure TGF2NECurve.Mul(P: TGF2NEPoint; K: TInteger; Jacobian: Boolean = false);
  const Selector : array[Boolean,Boolean] of Byte = ((0,1),(2,0));
  var
    PJX, PJY, PJZ, QpJX, QpJY, QpJZ, QmJY : TGF2Poly;
    Qp, Qm                                : TGF2NEPoint;
    H                                     : TInteger;
    i                                     : Longint;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Mul');

  //===== Handle special cases
  //----- If P = O then nothing to do
  if P.IsO then Exit;
  //----- If K = 0, return P.O
  if K.Is0 then
  begin
    P.SetO;
    Exit;
  end;
  //----- If K = 1, nothing to do
  if K.Is1 then Exit;
  //---- If K negative, (PX,PY)*(-K) = -(PX,PY)*K
  if K.Neg then
  begin
    Negate(P);
    if K.IsMinus1 then Exit;
  end;

  //===== Use algorithm described in IEEE-P1363 Annex A
  //----- Jacobian coordinates
  if Jacobian then
  begin
    PJX := nil;
    PJY := nil;
    PJZ := nil;
    QpJX := nil;
    QpJY := nil;
    QpJZ := nil;
    QmJY := nil;
    H := TInteger.Create(K);
    try
      //----- Convert from affine to Jacobian coordinates
      PJX := TGF2Poly.Create(P.fX);
      PJY := TGF2Poly.Create(P.fY);
      PJZ := TGF2Poly.Create;
      PJZ.Set1;
      QpJX := TGF2Poly.Create(PJX);
      QpJY := TGF2Poly.Create(PJY);
      QpJZ := TGF2Poly.Create(PJZ);
      QmJY := TGF2Poly.Create(PJY);
      fF.Add(QmJY,PJX);
      //----- Multiplication
      IMul(H,3);
      for i := H.BitSize-2 downto 1 do
      begin
        JDouble(PJX,PJY,PJZ);
        case Selector[H.Bit[i],K.Bit[i]] of
          1: JAdd(PJX,PJY,PJZ,QpJX,QmJY,QpJZ); //P := P - Q
          2: JAdd(PJX,PJY,PJZ,QpJX,QpJY,QpJZ); //P := P + Q
        end;
      end;
      //----- Convert to from Jacobian to affine coordinates
      JToAC(PJX,PJY,PJZ,P);
    finally
      H.Free;
      PJX.Free;
      PJY.Free;
      PJZ.Free;
      QpJX.Free;
      QpJY.Free;
      QpJZ.Free;
      QmJY.Free;
    end;
  end
  //----- Affine coordinates
  else
  begin
    Qp := nil;
    Qm := nil;
    H := TInteger.Create(K);
    try
      Qp := TGF2NEPoint.Create(P);
      Qm := TGF2NEPoint.Create(P);
      Negate(Qm);
      IMul(H,3);
      for i := H.BitSize-2 downto 1 do
      begin
        Double(P);
        case Selector[H.Bit[i],K.Bit[i]] of
          1: Add(P,Qm); //P := P - Q
          2: Add(P,Qp); //P := P + Q
        end;
      end;
    finally
      H.Free;
      Qp.Free;
      Qm.Free;
    end;
  end;
end;

//==============================================================================
// Negate
// <- P := -P  (if U = (X,Y) then -U = (X,Y+X))
//==============================================================================
procedure TGF2NECurve.Negate(P: TGF2NEPoint);
begin
  //----- If P = O or (PX = 0) then nothing to do
  //      (PX <> 0) => (P <> O), so no need to test PY
  with P do if not fX.Is0 then fF.Add(fY,fX);
end;

//==============================================================================
// Return TRUE iff P is on the curve
//==============================================================================
function TGF2NECurve.OnCurve(P: TGF2NEPoint): Boolean;
  var S, T : TGF2Poly;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('OnCurve');

  if P.IsO then Result := true
  else
  begin
    T := nil;
    S := TGF2Poly.Create(P.fX);
    try
      fF.Sqr(S);
      if fA.Is0 then //Spare 1 polynomial multiplication
      begin
        //----- S = X(X + Y) + Y
        fF.Add(S,P.fY);
        fF.Mul(S,P.fX);
        T := TGF2Poly.Create(P.fY);
      end
      else
      begin
        //----- S = X^3 + AX^2 + XY + Y^2
        T := TGF2Poly.Create(P.fX);
        fF.Add(T,fA);
        fF.Mul(S,T);
        GF2Set(T,P.fY);
        fF.Mul(T,P.fX);
        fF.Add(S,T);
        GF2Set(T,P.fY);
      end;
      fF.Sqr(T);
      fF.Add(S,T);
      //----- S = B ?
      Result := GF2Equ(S,fB);
    finally
      S.Free;
      T.Free;
    end;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,SAVEFLAGS);
  try
    F.WriteBuffer(ECHeaderFile,SizeOf(ECHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.SaveToStream(Stream: TStream);
begin
  Stream.WriteBuffer(fStatus,SizeOf(fStatus));
  fF.SaveToStream(Stream);
  fA.SaveToStream(Stream);
  fB.SaveToStream(Stream);
  fC.SaveToStream(Stream);
end;

//==============================================================================
// Overloaded
// Set with "degree" array (See GF2SetDegs)
//==============================================================================
procedure TGF2NECurve.SetCoefs(const A, B: array of Longint);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseInvalidArg(ClassName + '.SetCoefs','Invalid degrees');
  end;
  //--------------------------------------------------------------------------
  var Z : TGF2Poly;
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('SetCoefs');

  Z := TGF2Poly.Create;
  try
    //----- Set B
    try
      GF2SetDegs(Z,B);
    except
      on EHITInvalidArg do Error;
    end;
    fF.Reduce(Z);
    //----- Check
    if Z.Is0 then Error;
    GF2Set(fB,Z);
    //----- Set A
    try
      GF2SetDegs(Z,A);
    except
      on EHITInvalidArg do Error;
    end;
    fF.Reduce(Z);
    GF2Set(fA,Z);
  finally
    Z.Free;
  end;

  GF2Set(fC,fB);
  fF.RepeatedSqr(fC,fF.ModulusDegree-2);

  fStatus := csHasCoefs;
end;

//==============================================================================
// Overloaded
// Set A and B
// If new coefs are not valid, the old ones (if any) are not changed
//==============================================================================
procedure TGF2NECurve.SetCoefs(A, B: TGF2Poly);
  var Z : TGF2Poly;
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('SetCoefs');

  Z := TGF2Poly.Create(B);
  try
    fF.Reduce(Z);
    //----- Check
    if Z.Is0 then
      HITRaiseInvalidArg(ClassName + '.SetCoefs','Invalid B coefficient');
    GF2Set(fB,Z);
  finally
    Z.Free;
  end;
  if Assigned(A) then
  begin
    GF2Set(fA,A);
    fF.Reduce(fA);
  end
  else fA.Set0;

  GF2Set(fC,fB);
  fF.RepeatedSqr(fC,fF.ModulusDegree-2);

  fStatus := csHasCoefs;
end;

//==============================================================================
// Overloaded
// We must have Degrees = [dmax, ..., 0] and Degrees must be an irreducible
// polynomial over GF(2)
//
// If the new field is not valid, the old one (if any) is not changed
//==============================================================================
procedure TGF2NECurve.SetField(const Degrees: array of Longint; Check: Boolean);
  var Gf : TGF2N;
begin
  Gf := TGF2N.Create;
  try
    Gf.SetModulus(Degrees,Check);
  except
    Gf.Free;
    raise;
  end;
  //----- All is right, set the new field
  fF.Free;
  fF := Gf;
  fStatus := csHasField;
end;

//==============================================================================
// Overloaded
//==============================================================================
procedure TGF2NECurve.SetField(F: TGF2N);
begin
  if Assigned(F) then
  begin
    fStatus := csEmpty;
    fF.Copy(F);
    fStatus := csHasField;
  end;
end;

//==============================================================================
// Overloaded
// P must be an irreducible polynomial over GF(2)
// If the new field is not valid, the old one (if any) is not changed
//==============================================================================
procedure TGF2NECurve.SetField(P: TGF2Poly; Check: Boolean);
  var Gf : TGF2N;
begin
  Gf := TGF2N.Create;
  try
    Gf.SetModulus(P,Check);
  except
    Gf.Free;
    raise;
  end;
  //----- All is right, set the new field
  fF.Free;
  fF := Gf;
  fStatus := csHasField;
end;

//==============================================================================
// Generate random coefficients fA and fB
//==============================================================================
procedure TGF2NECurve.SetRndCoefs(BOnly: Boolean = false);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('SetRndCoefs');

  fStatus := csHasField;
  if BOnly then fA.Set0 else fF.Rnd(fA);
  repeat
    fF.Rnd(fB);
  until not fB.Is0;
  fStatus := csHasCoefs;
end;

//==============================================================================
// Sub
// -> P, Q, points
// <- P = P - Q
//==============================================================================
procedure TGF2NECurve.Sub(P, Q: TGF2NEPoint);
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Sub');

  if P = Q then P.SetO
  else
  begin
    fF.Add(P.fY,P.fX); //Negate P (if P = O then fX = 0, all is right)
    Add(P,Q);
    fF.Add(P.fY,P.fX);
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

