////////////////////////////////////////////////////////////////////////////////
// Elliptic Curves over GF(P) (with P prime > 3)
// Unit of the HIT package for DELPHI 5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit EllipticCurvesGFP;

//==============================================================================
{.$DEFINE DEBUG}
//------------------------------------------------------------------------------
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G-,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Classes, SysUtils, Integers;

//==============================================================================
// The elliptic curves over GF(P) used here are always
// Y^2 = X^3 + AX + B
// Such a curve is non singular if and only if
// ((4A^3 + 27B^2) <> 0 mod P) and (GCD(P,6) = 1)
//==============================================================================

type
  //===== Point of a curve Y^2 = X^3 + AX + B over GF(P)
  //      Point at infinity (X,Y,O) = (any,any,true)
  TGFPEPoint = class(THITObject)
  private
    fX : TInteger;
    fY : TInteger;
    fO : Boolean;
  public
    constructor Create; overload;
    constructor Create(P: TGFPEPoint); overload;
    destructor  Destroy; override;
    //-----
    procedure Copy(P: TGFPEPoint);
    function  Equ(P: TGFPEPoint): Boolean;
    function  IsO: Boolean;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure SetO;
    procedure SetXY(X, Y: TInteger);
    //-----
    property X: TInteger read fX;
    property Y: TInteger read fY;
  end;

  //===== Curve (Weierstrass form)
  //      Y^2 = X^3 + AX + B over GF(P)
  TGFPECStatus = (csEmpty, csHasModulus, csHasCoefs);
  TGFPECurve = class(THITObject)
  private
    fStatus: TGFPECStatus;
    fM: TInteger; //Modulus, prime > 3
    fA: TInteger; //Coef, integer over GF(M)
    fB: TInteger; //Coef, integer over GF(M)
    //-----
    procedure ErrorNoCoefs(const Caller: string);
    procedure ErrorNoModulus(const Caller: string);
    procedure JAdd(X0,Y0,Z0, X1,Y1,Z1: TInteger);
    procedure JDouble(X,Y,Z: TInteger);
    procedure JToAC(JX,JY,JZ: TInteger; P: TGFPEPoint);
  public
    class function CheckCurve(A, B, M: TInteger): Boolean;
    class function CheckModulus(M: TInteger): Boolean;
    //-----
    constructor Create; overload;
    constructor Create(C: TGFPECurve); overload;
    destructor  Destroy; override;
    //-----
    procedure Add(P, Q: TGFPEPoint);
    procedure Copy(C: TGFPECurve);
    procedure Double(P: TGFPEPoint);
    procedure GetCoefs(A, B: TInteger);
    procedure GetModulus(M: TInteger);
    procedure GetRndPoint(P: TGFPEPoint);
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure Mul(P: TGFPEPoint; K: TInteger; Jacobian: Boolean = false);
    procedure Negate(P: TGFPEPoint);
    function  OnCurve(P: TGFPEPoint): Boolean;
    procedure SetCoefs(A, B: TInteger);
    procedure SetModulus(M: TInteger);
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure SetRndCoefs;
    procedure Sub(P, Q: TGFPEPoint);
    //-----
    property  Status: TGFPECStatus read fStatus;
  end;


implementation
//==============================================================================
uses HIT_Hash, Integers_Primes;

////////////////////////////////////////////////////////////////////////////////
// CONSTS, VARS, TYPES, ETC.
////////////////////////////////////////////////////////////////////////////////

const
  //----- File flags
  LOADFLAGS = fmOpenRead or fmShareDenyWrite;
  SAVEFLAGS = fmCreate or fmShareExclusive;

type
  TEPointHeaderFile = array[0..23] of AnsiChar;
  TECurveHeaderFile = array[0..23] of AnsiChar;

const
  //----- Header of "point" file
  EPointHeaderFile : TEPointHeaderFile = 'HIT EC over GFP - Point ';
  //----- Header of "curve" file
  ECurveHeaderFile : TECurveHeaderFile = 'HIT EC over GFP - Curve ';


////////////////////////////////////////////////////////////////////////////////
// TGFPEPoint CLASS
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
//==============================================================================
procedure TGFPEPoint.Copy(P: TGFPEPoint);
begin
  ISet(fX,P.fX);
  ISet(fY,P.fY);
  fO := P.fO;
end;

//==============================================================================
//==============================================================================
constructor TGFPEPoint.Create;
begin
  fX := TInteger.Create;
  fY := TInteger.Create;
  fO := true;
end;

//==============================================================================
//==============================================================================
constructor TGFPEPoint.Create(P: TGFPEPoint);
begin
  fX := TInteger.Create(P.fX);
  fY := TInteger.Create(P.fY);
  fO := P.fO;
end;

//==============================================================================
//==============================================================================
destructor TGFPEPoint.Destroy;
begin
  fX.Free;
  fY.Free;
end;

//==============================================================================
// Result := TRUE iff Self = P
//==============================================================================
function TGFPEPoint.Equ(P: TGFPEPoint): Boolean;
begin
  if fO then Result := P.fO
        else Result := (not P.fO) and IEqu(fX,P.fX) and IEqu(fY,P.fY);
end;

//==============================================================================
// Return TRUE iff the point is the "point at infinity"
//==============================================================================
function TGFPEPoint.IsO: Boolean;
begin
  Result := fO;
end;

//==============================================================================
// Load Self from a file
//==============================================================================
procedure TGFPEPoint.LoadFromFile(const FileName: string);
  //--------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------
  var
    F : TFileStream;
    H : TEPointHeaderFile;
begin
  F := TFileStream.Create(FileName,LOADFLAGS);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> EPointHeaderFile then Error;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGFPEPoint.LoadFromStream(Stream: TStream);
begin
  fX.LoadFromStream(Stream);
  fY.LoadFromStream(Stream);
  Stream.ReadBuffer(fO,SizeOf(fO));
end;

//==============================================================================
//==============================================================================
procedure TGFPEPoint.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,SAVEFLAGS);
  try
    F.WriteBuffer(EPointHeaderFile,SizeOf(EPointHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGFPEPoint.SaveToStream(Stream: TStream);
begin
  fX.SaveToStream(Stream);
  fY.SaveToStream(Stream);
  Stream.WriteBuffer(fO,SizeOf(fO));
end;

//==============================================================================
//==============================================================================
procedure TGFPEPoint.SetO;
begin
  fX.Set0; //Cleaner
  fY.Set0;
  fO := true;
end;

//==============================================================================
//==============================================================================
procedure TGFPEPoint.SetXY(X, Y: TInteger);
begin
  ISet(fX,X);
  ISet(fY,Y);
  fO := false;
end;

////////////////////////////////////////////////////////////////////////////////
// LOCAL ROUTINES
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Check whether the coefs A,B,M define a regular elliptic curve or not
// -> A, B
// -> M, such that M > 3 and GCD(M,6) = 1
// <- Return TRUE iff (4A^3 + 27B^2) <> 0 mod M
//------------------------------------------------------------------------------
function CheckAB(A,B,M: TInteger): Boolean;
  var D, T : TInteger;
begin
  T := nil;
  D := TInteger.Create(A);
  try
    ISqr(D);
    IMod(D,M);
    IMul(D,A);
    IShl(D,2);
    T := TInteger.Create(B);
    ISqr(T);
    IMod(T,M);
    IMul(T,27);
    IAdd(D,T);
    IMod(D,M);
    Result := not D.Is0;
  finally
    D.Free;
    T.Free;
  end;
end;

//------------------------------------------------------------------------------
// Return Z := (X^3 + AX + B) mod M with 0 <= Z < M
// -> X, coordinate of a point of E(A,B)/M
// -> (A,B,M), elliptic curve
// <- Z
//------------------------------------------------------------------------------
procedure EvalY2(Z, X,A,B,M: TInteger);
begin
  ISet(Z,X);
  ISqr(Z);
  IMod(Z,M);
  IAdd(Z,A);
  IMul(Z,X);
  IMod(Z,M);
  IAdd(Z,B);
  IMod(Z,M);
end;

////////////////////////////////////////////////////////////////////////////////
// TGFPECurve CLASS
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Check whether the coefs A,B,M define a regular elliptic curve or not
// -> A,B,M
// <- Return TRUE iff ((4A^3 + 27B^2) <> 0 mod M) and (M > 3) and (GCD(M,6) = 1)
//==============================================================================
class function TGFPECurve.CheckCurve(A, B, M: TInteger): Boolean;
begin
  Result := CheckModulus(M) and CheckAB(A,B,M);
end;

//==============================================================================
// Return TRUE iff (M > 3) and (GCD(M,6) = 1)
//==============================================================================
class function TGFPECurve.CheckModulus(M: TInteger): Boolean;
begin
  Result := M.Odd and (ICmp(M,3) > 0) and (IMod(M,3) <> 0);
end;

//==============================================================================
// Add
// -> P, Q, points of E(A,B)/M
// <- P = P + Q
//==============================================================================
procedure TGFPECurve.Add(P, Q: TGFPEPoint);
  var S, T : TInteger;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Add');

  //----- If Q = Identity, there is nothing to do
  if Q.fO then Exit;
  //----- If P = Identity then P := Q
  if P.fO then
  begin
    P.Copy(Q);
    Exit;
  end;
  //----- If P = Q or -Q (Handle also the case "Add(P,P)")
  if IEqu(P.fX,Q.fX) then //
  begin
    //----- If (P = -Q) then P := Identity else P := 2P
    if P.fY.Is0 or not IEqu(P.fY,Q.fY) then P.SetO else Double(P);
    Exit;
  end;

  S := nil;
  T := TInteger.Create(P.fX);
  with P do
  try
    //----- T := (PX - QX)^(-1) mod M
    //      The inversion may fail if M is not prime.
    ISub(T,Q.fX);
    if not IInvMod(T,fM) then
      HITError(ClassName + '.Add'#13#10'Failed inversion');
    //----- S := (PY - QY) mod M
    S := TInteger.Create(fY);
    ISub(S,Q.fY);
    if S.Neg then IAdd(S,fM);
    //----- S := (PY - QY)/(PX - QX) mod M
    IMul(S,T);
    IMod(S,fM);
    //----- PX := (S - PX - QX) mod M
    ISet(T,S);
    ISwp(fX,T); //PX := S, T := iv(PX) (initial value of PX)
    ISqr(fX);
    IMod(fX,fM);
    ISub(fX,T);
    ISub(fX,Q.fX);
    while fX.Neg do IAdd(fX,fM); //Add at most twice
    //----- PY := ((PX - X)S - PY) mod M
    ISwp(fY,T); //PY := iv(PX), T := iv(PY)
    ISub(fY,fX);
    if fY.Neg then IAdd(fY,fM);
    IMul(fY,S);
    IMod(fY,fM);
    ISub(fY,T);
    if fY.Neg then IAdd(fY,fM);
  finally
    S.Free;
    T.Free;
  end;
end;

//==============================================================================
// Set Self equal to C
//==============================================================================
procedure TGFPECurve.Copy(C: TGFPECurve);
begin
  ISet(fM,C.fM);
  ISet(fA,C.fA);
  ISet(fB,C.fB);
  fStatus := C.fStatus;
end;

//==============================================================================
//==============================================================================
constructor TGFPECurve.Create;
begin
  fM := TInteger.Create;
  fA := TInteger.Create;
  fB := TInteger.Create;
end;

//==============================================================================
//==============================================================================
constructor TGFPECurve.Create(C: TGFPECurve);
begin
  fM := TInteger.Create(C.fM);
  fA := TInteger.Create(C.fA);
  fB := TInteger.Create(C.fB);
  fStatus := C.fStatus;
end;

//==============================================================================
//==============================================================================
destructor TGFPECurve.Destroy;
begin
  fM.Free;
  fA.Free;
  fB.Free;
end;

//==============================================================================
// Double
// -> P, point of E(A,B)/M
// <- P = P x 2
//==============================================================================
procedure TGFPECurve.Double(P: TGFPEPoint);
  var S, T : TInteger;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Double');

  with P do
  begin
    //----- If P = Identity, there is nothing to do
    if fO then Exit;
    S := nil;
    T := TInteger.Create(fY);
    try
      //----- Si (Y + Y) = 0 mod M then return P.Identity = true
      //      When P is an odd prime ((Y + Y) = 0) -> (Y = 0)
      IShl(T);
      IMod(T,fM);
      if T.Is0 then SetO //P := O
      else
      begin
        //----- T := (2Y)^(-1) mod M
        //      The inversion may fail if M is not prime.
        if not IInvMod(T,fM) then
          HITError(ClassName + '.Double'#13#10'Failed inversion');
        //----- S := (3 X + A) mod M
        S := TInteger.Create(fX);
        ISqr(S);
        IMod(S,fM);
        IMul(S,3);
        IAdd(S,fA);
        IMod(S,fM);
        //----- S := (3 X + A)/(2Y) mod M
        IMul(S,T);
        IMod(S,fM);
        //----- X := (S - 2X) mod M
        ISet(T,S);
        ISwp(fX,T); //X := S, T := iv(X) ("iv(X)" means "initial value of X")
        ISqr(fX);
        IMod(fX,fM);
        ISub(fX,T);
        if fX.Neg then IAdd(fX,fM);
        ISub(fX,T);
        if fX.Neg then IAdd(fX,fM);
        //----- Y := ((iv(X) - X)S - Y) mod M
        ISwp(fY,T); //Y := iv(X) and T := iv(Y)
        ISub(fY,fX);
        if fY.Neg then IAdd(fY,fM);
        IMul(fY,S);
        IMod(fY,fM);
        ISub(fY,T);
        if fY.Neg then IAdd(fY,fM);
      end;
    finally
      S.Free;
      T.Free;
    end;
  end;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure TGFPECurve.ErrorNoCoefs(const Caller: string);
begin
  HITRaiseInvalidCall(ClassName + '.' + Caller,'Curve without coefs');
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure TGFPECurve.ErrorNoModulus(const Caller: string);
begin
  HITRaiseInvalidCall(ClassName + '.' + Caller,'Curve without modulus');
end;

//==============================================================================
//==============================================================================
procedure TGFPECurve.GetCoefs(A, B: TInteger);
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('GetCoefs');

  if Assigned(A) then ISet(A,fA);
  if Assigned(B) then ISet(B,fB);
end;

//==============================================================================
// Return the curve modulus
//==============================================================================
procedure TGFPECurve.GetModulus(M: TInteger);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoModulus('GetModulus');

  ISet(M,fM);
end;

//=======================================================================
// Generate a random point of an elliptic curve
// <- (X,Y,false), random point in the group E(A,B)/M
//=======================================================================
procedure TGFPECurve.GetRndPoint(P: TGFPEPoint);
  var s1, s2 : Longint;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('GetRndPoint');

  with P do
  begin
    fO := false;
    s1 := fM.BitSize + 1;
    s2 := s1 + s1;
    repeat
      IRnd(fX,s2);
      IH160(fX,s1,20);
      IMod(fX,fM);
      EvalY2(fY,fX,fA,fB,fM);
    until IJacobi(fY,fM) = 1;
    ISqrRootMod(fY,fM);
    //----- Toss up for Y or -Y (mod M)
    if not fY.Is0 then
      if Odd(IRnd32) then ISubr(fY,fM);
  end;
end;

//------------------------------------------------------------------------------
// Adding in Jacobian coordinates
// -> (X0,Y0,Z0) and (X1,Y1,Z1), points of the curve
// <- (X0,Y0,Z0) = (X0,Y0,Z0) + (X1,Y1,Z1)
//------------------------------------------------------------------------------
procedure TGFPECurve.JAdd(X0,Y0,Z0, X1,Y1,Z1: TInteger);
  var R, S, T : TInteger;
begin
  R := nil;
  S := nil;
  T := nil;
  try
    R := TInteger.Create(X1);
    S := TInteger.Create(Y1);
    T := TInteger.Create;
    if not Z1.Is1 then
    begin
      ISet(T,Z1); ISqr(T); IMod(T,fM);
      IMul(X0,T); IMod(X0,fM);
      IMul(T,Z1); IMod(T,fM);
      IMul(Y0,T); IMod(Y0,fM);
    end;
    if not Z0.Is1 then
    begin
      ISet(T,Z0); ISqr(T); IMod(T,fM);
      IMul(R,T); IMod(R,fM);
      IMul(T,Z0); IMod(T,fM);
      IMul(S,T); IMod(S,fM);
    end;
    ISub(R,X0); R.Negate; if R.Neg then IAdd(R,fM);
    ISub(S,Y0); S.Negate; if S.Neg then IAdd(S,fM);
    if R.Is0 then
    begin
      if S.Is0 then
      begin
        X0.Set0;
        Y0.Set0;
      end
      else
      begin
        X0.Set1;
        Y0.Set1;
      end;
      Z0.Set0;
      Exit;
    end;
    IShl(X0); ISub(X0,R);
    if X0.Neg then IAdd(X0,fM) else if ICmp(X0,fM) > 0 then ISub(X0,fM);
    IShl(Y0); ISub(Y0,S);
    if Y0.Neg then IAdd(Y0,fM) else if ICmp(Y0,fM) > 0 then ISub(Y0,fM);
    if not Z1.Is1 then
    begin
      IMul(Z0,Z1); IMod(Z0,fM);
    end;
    IMul(Z0,R); IMod(Z0,fM);
    ISet(T,R); ISqr(T); IMod(T,fM);
    IMul(R,T); IMod(R,fM);
    IMul(T,X0); IMod(T,fM);
    ISet(X0,S); ISqr(X0); IMod(X0,fM);
    ISub(X0,T); if X0.Neg then IAdd(X0,fM);
    ISub(T,X0); if T.Neg then IAdd(T,fM);
    ISub(T,X0); if T.Neg then IAdd(T,fM);
    IMul(S,T); IMod(S,fM);
    IMul(R,Y0); IMod(R,fM);
    ISet(Y0,S); ISub(Y0,R); if Y0.Neg then IAdd(Y0,fM);
    if Y0.Odd then IAdd(Y0,fM);
    IShr(Y0);
  finally
    R.Free;
    S.Free;
    T.Free;
  end;
end;

//------------------------------------------------------------------------------
// Doubling in Jacobian coordinates
// -> (X,Y,Z) point of the curve
// <- (X,Y,Z) = 2(X,Y,Z)
//------------------------------------------------------------------------------
procedure TGFPECurve.JDouble(X,Y,Z: TInteger);
  var S, T : TInteger;
begin
  //----- Handle special case
  if Y.Is0 or Z.Is0 then
  begin
    X.Set1;
    Y.Set1;
    Z.Set0;
    Exit;
  end;

  S := nil;
  T := nil;
  try
    S := TInteger.Create(X);
    T := TInteger.Create(Z);
    ISqr(T); IMod(T,fM);
    ISqr(T); IMod(T,fM);
    IMul(T,fA); IMod(T,fM);
    ISqr(S); IMod(S,fM);
    IMul(S,3); IAdd(S,T); IMod(S,fM);
    IShl(Z); IMul(Z,Y); IMod(Z,fM);
    ISqr(Y); IMod(Y,fM);
    ISet(T,X); IShl(T,2); IMul(T,Y); IMod(T,fM);
    ISet(X,S); ISqr(X); IMod(X,fM);
    ISub(X,T); if X.Neg then IAdd(X,fM);
    ISub(X,T); if X.Neg then IAdd(X,fM);
    IShl(Y); ISqr(Y); IMod(Y,fM);
    IShl(Y); if ICmp(Y,fM) > 0 then ISub(Y,fM);
    ISub(T,X); IMul(T,S); IMod(T,fM);
    ISub(Y,T); Y.Negate; if Y.Neg then IAdd(Y,fM);
  finally
    S.Free;
    T.Free;
  end;
end;

//------------------------------------------------------------------------------
// Conversion: Jacobian coordinates -> Affine coordinates
//------------------------------------------------------------------------------
procedure TGFPECurve.JToAC(JX,JY,JZ: TInteger; P: TGFPEPoint);
  var U, V : TInteger;
begin
  if JZ.Is0 then
  begin
    P.fX.Set0;
    P.fY.Set0;
    P.fO := true;
  end
  else
  begin
    V := nil;
    U := TInteger.Create(JZ);
    try
      IInvMod(U,fM);
      V := TInteger.Create(U);
      ISqr(U); IMod(U,fM);
      ISet(P.fX,JX); IMul(P.fX,U); IMod(P.fX,fM); //AX := JX / JZ^2
      IMul(U,V); IMod(U,fM);
      ISet(P.fY,JY); IMul(P.fY,U); IMod(P.fY,fM); //AY := JY / JZ^3
    finally
      U.Free;
      V.Free;
    end;
    P.fO := false;
  end;
end;

//==============================================================================
// Load Self from a file
//==============================================================================
procedure TGFPECurve.LoadFromFile(const FileName: string);
  //--------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------
  var
    F : TFileStream;
    H : TECurveHeaderFile;
begin
  F := TFileStream.Create(FileName,LOADFLAGS);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> ECurveHeaderFile then Error;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
// Load Self from a stream
//==============================================================================
procedure TGFPECurve.LoadFromStream(Stream: TStream);
begin
  Stream.ReadBuffer(fStatus,SizeOf(fStatus));
  fM.LoadFromStream(Stream);
  fA.LoadFromStream(Stream);
  fB.LoadFromStream(Stream);
end;

//==============================================================================
// Mul
// -> P, point of E(A,B)/M
// -> K, integer
// <- P = P x K
//==============================================================================
procedure TGFPECurve.Mul(P: TGFPEPoint; K: TInteger; Jacobian: Boolean = false);
  const Selector : array[Boolean,Boolean] of Byte = ((0,1),(2,0));
  var
    PJX, PJY, PJZ, QpJX, QpJY, QpJZ, QmJY : TInteger;
    Qm, Qp                                : TGFPEPoint;
    H                                     : TInteger;
    i                                     : Longint;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Mul');

  //----- If P = Identity then nothing to do
  if P.fO then Exit;
  //----- If K = 0, return P = Identity
  if K.Is0 then
  begin
    P.SetO;
    Exit;
  end;
  //----- If K = 1, nothing to do
  if K.Is1 then Exit;
  //---- If K negative, (PX,PY)*(-K) = (PX,-PY)*K
  if K.Neg then
  begin
    if not P.fY.Is0 then ISubr(P.fY,fM);
    if K.IsMinus1 then Exit;
  end;

  //===== Use algorithm described in IEEE-P1363/D8 Annex A
  if Jacobian then
  begin
    PJX := nil;
    PJY := nil;
    PJZ := nil;
    QpJX := nil;
    QpJY := nil;
    QpJZ := nil;
    QmJY := nil;
    H := TInteger.Create(K);
    try
      //----- Convert to from affine to Jacobian coordinates
      PJX := TInteger.Create(P.fX);
      PJY := TInteger.Create(P.fY);
      PJZ := TInteger.Create;
      PJZ.Set1;
      QpJX := TInteger.Create(PJX);
      QpJY := TInteger.Create(PJY);
      QpJZ := TInteger.Create(PJZ);
      QmJY := TInteger.Create(PJY);
      if not QmJY.Is0 then ISubr(QmJY,fM);

      //----- Multiply
      IMul(H,3);
      for i := H.BitSize-2 downto 1 do
      begin
        JDouble(PJX,PJY,PJZ);
        case Selector[H.Bit[i],K.Bit[i]] of
          1: JAdd(PJX,PJY,PJZ,QpJX,QmJY,QpJZ); //P := P - Q
          2: JAdd(PJX,PJY,PJZ,QpJX,QpJY,QpJZ); //P := P + Q
        end;
      end;

      //----- Convert to from Jacobian to affine coordinates
      JToAC(PJX,PJY,PJZ, P);
    finally
      H.Free;
      PJX.Free;
      PJY.Free;
      PJZ.Free;
      QpJX.Free;
      QpJY.Free;
      QpJZ.Free;
      QmJY.Free;
    end;
  end
  else
  begin
    Qp := nil;
    Qm := nil;
    H := TInteger.Create(K);
    try
      Qp := TGFPEPoint.Create(P);
      Qm := TGFPEPoint.Create(P);
      Negate(Qm);
      IMul(H,3);
      for i := H.BitSize-2 downto 1 do
      begin
        Double(P);
        case Selector[H.Bit[i],K.Bit[i]] of
          1: Add(P,Qm); //P := P - Qp
          2: Add(P,Qp); //P := P + Qp
        end;
      end;
    finally
      Qp.Free;
      Qm.Free;
      H.Free;
    end;
  end;
end;

//==============================================================================
// Negate
// -> P, point of E(A,B)/M
// <- P = -P
//==============================================================================
procedure TGFPECurve.Negate(P: TGFPEPoint);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoModulus('Negate');

  //----- If P = Identity or (PY = 0) then nothing to do
  with P do if not (fO or fY.Is0) then ISubr(fY,fM);
end;

//==============================================================================
// Return TRUE iff P is on the curve (A,B)/M
// -> P, candidate
//==============================================================================
function TGFPECurve.OnCurve(P: TGFPEPoint): Boolean;
  var S, T : TInteger;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('OnCurve');

  Result := P.fO;
  if Result then Exit;
  T := nil;
  S := TInteger.Create;
  try
    //----- S := X^3 + AX + B mod M
    EvalY2(S,P.fX,fA,fB,fM);
    //----- T := Y
    T := TInteger.Create(P.fY);
    ISqr(T);
    IMod(T,fM);
    //----- S = T ?
    Result := IEqu(S,T);
  finally
    S.Free;
    T.Free;
  end;
end;

//==============================================================================
// Save Self to a file
//==============================================================================
procedure TGFPECurve.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,SAVEFLAGS);
  try
    F.WriteBuffer(ECurveHeaderFile,SizeOf(ECurveHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Save Self to a stream
//==============================================================================
procedure TGFPECurve.SaveToStream(Stream: TStream);
begin
  Stream.WriteBuffer(fStatus,SizeOf(fStatus));
  fM.SaveToStream(Stream);
  fA.SaveToStream(Stream);
  fB.SaveToStream(Stream);
end;

//==============================================================================
//==============================================================================
procedure TGFPECurve.SetCoefs(A, B: TInteger);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoModulus('SetCoefs');
  if not CheckCurve(A,B,fM) then
    HITRaiseInvalidArg(ClassName+'.SetCoefs','Invalid coefficients');

  ISet(fA,A);
  ISet(fB,B);
  fStatus := csHasCoefs;
end;

//==============================================================================
//==============================================================================
procedure TGFPECurve.SetModulus(M: TInteger);
begin
  //----- Check
  if not CheckModulus(M) then
    HITRaiseInvalidArg(ClassName+'.SetModulus','Invalid modulus');

  ISet(fM,M);
  fStatus := csHasModulus;
end;

//==============================================================================
// Generate random coefficients of an elliptic curve
// <- (fA,fB), random coefficients
//==============================================================================
procedure TGFPECurve.SetRndCoefs;
  var s1, s2 : Longint;
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoModulus('SetRndCoefs');

  s1 := fM.BitSize + 1;
  s2 := s1 + s1;
  repeat
    IRnd(fA,s2);
    IH160(fA,s1,10);
    IMod(fA,fM);
    IRnd(fB,s2);
    IH160(fB,s1,11);
    IMod(fB,fM);
  until CheckAB(fA,fB,fM);
  fStatus := csHasCoefs;
end;

//==============================================================================
// Sub
// -> P, Q, points of E(A,B)/M
// <- P = P - Q
//==============================================================================
procedure TGFPECurve.Sub(P, Q: TGFPEPoint);
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Sub');

  Negate(P);
  Add(P,Q);
  Negate(P);
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

