////////////////////////////////////////////////////////////////////////////////
// Lucas Primality Test
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit LucasTest;

//==============================================================================
{.$DEFINE DEBUG}
//------------------------------------------------------------------------------
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers;

//------------------------------------------------------------------------------
// Lucas primality test
// N must be odd
// F must contain all the prime odd factors (without exponent) of N+1
//------------------------------------------------------------------------------

function LTest(N: TInteger; const F: array of Longword): Integer; overload;
function LTest(N: TInteger; const F: array of TInteger): Integer; overload;

implementation
//==============================================================================
uses Integers_Primes;

////////////////////////////////////////////////////////////////////////////////
// MODULAR LUCAS SEQUENCE
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Calculate the V[k] element modulu N of a Lucas sequence
// -> K >= 0
// -> (P,Q), coefficients of the sequence, P*Q <> 0
// -> N, modulus, N >= 3 and N odd
// <- V := V[K] mod N
//==============================================================================
procedure ILucasMod(V, K, P, Q, N: TInteger);
  var
    M, T, W, X : TInteger;
    i          : Longint;
begin
  //----- if K = 0 then V := 2
  { In this unit K is always different from 0
  if K.Is0 then
  begin
    ISet(V,2);
    Exit;
  end;
  }
  //----- V := P
  ISet(V,P);
  IMod(V,N);
  //----- If K = 1, nothing more to do
  if K.Is1 then Exit;

  //----- K >= 2
  M := nil;
  T := nil;
  W := nil;
  X := nil;
  try
    M := TInteger.Create(Q);
    T := TInteger.Create(Q);
    W := TInteger.Create(P);
    X := TInteger.Create;
    //----- W = P - 2Q
    IShl(T);
    ISqr(W);
    IMod(W,N);
    ISub(W,T);
    if W.Neg then IAdd(W,N);
    //===== Loop
    for i := K.BitSize-2 downto 1 do
      if K.Bit[i] then
      begin
        //----- V = VW - PM
        IMul(V,W);
        IMod(V,N);
        ISet(T,P);
        IMul(T,M);
        IMod(T,N);
        ISub(V,T);
        if V.Neg then IAdd(V,N);
        //----- W = W - 2MQ
        ISqr(W);
        IMod(W,N);
        ISet(T,M);
        IShl(T);
        IMul(T,Q);
        IMod(T,N);
        ISub(W,T);
        if W.Neg then IAdd(W,N);
        //----- M := MQ
        ISqr(M);
        IMul(M,Q);
        IMod(M,N);
        if M.Neg then IAdd(M,N);
      end
      else
      begin
        //----- W := VW - PM
        IMul(W,V);
        IMod(W,N);
        ISet(T,P);
        IMul(T,M);
        IMod(T,N);
        ISub(W,T);
        if W.Neg then IAdd(W,N);
        //----- V = V - 2M
        ISqr(V);
        IMod(V,N);
        ISet(T,M);
        IShl(T);
        IMod(T,N);
        ISub(V,T);
        if V.Neg then IAdd(V,N);
        //----- M = M
        ISqr(M);
        IMod(M,N);
      end;
    //===== Trailing bit of K
    if K.Odd then
    begin
      //----- V = VW - PM
      IMul(V,W);
      IMod(V,N);
      IMul(M,P);
      IMod(M,N);
      ISub(V,M);
      if V.Neg then IAdd(V,N);
    end
    else
    begin
      //----- V = V - 2M
      ISqr(V);
      IMod(V,N);
      IShl(M);
      ISub(V,M);
      if V.Neg then IAdd(V,N);
    end;
  finally
    M.Free;
    T.Free;
    W.Free;
    X.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// LUCAS TESTS
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Generate random P and Q such that
// Jacobi(Q,N) = -1, Jacobi(P-4Q,N) = -1 and GCD(2Q(P-4Q),N) = 1
// N must be odd
//==============================================================================
procedure GetRandomPQ(P, Q, N: TInteger);
  var
    D, E : TInteger;
    bsz  : Longint;
begin
  D := nil;
  E := nil;
  try
    D := TInteger.Create;
    E := TInteger.Create;
    P.Set0;
    Q.Set0;
    //----- Q, random such that Jacobi(Q,N) = -1 and GCD(Q,N) = 1
    bsz := N.BitSize;
    IRnd(Q,Succ(bsz));
    repeat
      IInc(Q);
      IMod(Q,N);
    until IJacobi(Q,N) = -1; //if TRUE then IGCD1(Q,N) = TRUE
    //----- P, random such that Jacobi(P-4Q,N) = -1 and GCD((P-4Q),N) = 1
    IRnd(P,Succ(bsz));
    IMod(P,N);
    ISet(E,Q);
    IShl(E,2);
    IMod(E,N); //E = 4Q mod N
    ISet(D,P);
    ISqr(D);
    IMod(D,N);
    ISub(D,E);
    if D.Neg then IAdd(D,N); //D := (P-4Q) mod N
    if IJacobi(D,N) = -1 then Exit; //if TRUE then IGCD1(D,N) = TRUE
    ISet(E,P);
    IShl(E);
    IDec(E);
    IMod(E,N); //E := 2P+1 mod N
    repeat
      IInc(P);
      IMod(P,N); //P := (P+1) mod N
      IAdd(E,2); //E = 2P
      IAdd(D,E);
      IMod(D,N);
    until IJacobi(D,N) = -1; //if TRUE then IGCD1(D,N) = TRUE
  finally
    D.Free;
    E.Free;
  end;
end;

//==============================================================================
// Lucas primality test
// -> N must be odd
// -> F must contain all the prime odd factors (without exponent) of N+1
// <- -1  Exception, error, etc.
// <-  0  N is composite
// <-  1  N is prime
//==============================================================================
function LTest(N: TInteger; const F: array of TInteger): Integer;
  label NEWTRY;
  var
    K, P, Q, V : TInteger;
    i          : Longint;
begin
  Result := -1;
  if High(F) < 0 then Exit;
  //----- Test the primality of N
  K := nil;
  P := nil;
  Q := nil;
  V := nil;
  try
    K := TInteger.Create;
    P := TInteger.Create;
    Q := TInteger.Create;
    V := TInteger.Create;
  NEWTRY:
    GetRandomPQ(P,Q,N);
    //----- Calculate V[K] with K = (N+1)/2
    ISet(K,N);
    IInc(K);
    IShr(K);
    //----- Calculate the V[k] element of a Lucas sequence (modulo N)
    ILucasMod(V,K,P,Q,N);
    //----- If V[(N+1)/2] = 0, N might be prime
    if V.Is0 then
    begin
      for i := 0 to High(F) do
      begin
        //----- Calculate V[K] with K = (N+1)/(2*F[i])
        ISet(K,N);
        IInc(K);
        IShr(K);
        IDiv(K,F[i]);
        ILucasMod(V,K,P,Q,N);
        //----- If V[(N+1)/(2*F[i])] = 0, we cannot conclude.
        //      We have to retry with other values P and Q
        if V.Is0 then goto NEWTRY;
      end;
      //----- N is prime
      Result := 1;
    end
    //----- If V[(N+1)/2] <> 0 then N is composite
    else Result := 0;
  finally
    K.Free;
    P.Free;
    Q.Free;
    V.Free;
  end;
end;

//==============================================================================
//==============================================================================
function LTest(N: TInteger; const F: array of Longword): Integer;
  label NEXT;
  var
    K, P, Q, V : TInteger;
    i          : Longint;
begin
  Result := -1;
  if High(F) < 0 then Exit;
  K := nil;
  P := nil;
  Q := nil;
  V := nil;
  try
    K := TInteger.Create;
    P := TInteger.Create;
    Q := TInteger.Create;
    V := TInteger.Create;
  NEXT:
    GetRandomPQ(P,Q,N);
    ISet(K,N);
    IInc(K);
    IShr(K);
    //----- Calculate the V[k] element of a Lucas sequence (modulo N)
    ILucasMod(V,K,P,Q,N);
    if V.Is0 then
    begin
      for i := 0 to High(F) do
      begin
        ISet(K,N);
        IInc(K);
        IShr(K);
        IDiv(K,F[i]);
        ILucasMod(V,K,P,Q,N);
        if V.Is0 then goto NEXT;
      end;
      Result := 1;
    end
    else Result := 0;
  finally
    K.Free;
    P.Free;
    Q.Free;
    V.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

