////////////////////////////////////////////////////////////////////////////////
// Rabin PSS - Probabilistic Signature Scheme
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 1999, Marcel Martin. All rights reserved.
//------------------------------------------------------------------------------
// For theoretical explanations see
// "The Exact Security of Digital Signatures  How to Sign with RSA and Rabin"
// by Mihir Bellare and Phillip Rogaway in
// Advances in Cryptology  EUROCRYPT'96, Springer-Verlag, Berlin 1996.
// Also downloadable from:
// http://www-cse.ucsd.edu/users/mihir/papers/exactsigs.html
//------------------------------------------------------------------------------
// Warning (do not delete, please)
//
// The PSS scheme is patent-pending in the USA,
// for details contact Mat Grell at the
// Office of Technology Transfer
// mathew.grell@ucop.edu or (510) 587-6035.
// The Regents' of the University of California do not CONFER BY IMPLICATION,
// ESTOPPEL, OR OTHERWISE ANY LICENSE OR RIGHTS UNDER ANY PATENTS BY PERMITTING
// THIS PRE-PATENT USE.
// Conditions for licensing, if adopted as a standard by IEEE, are posted at
// http://grouper.ieee.org/groups/1363/letters/UC.html
//
////////////////////////////////////////////////////////////////////////////////
unit RabinPSS;

//==============================================================================
{.$DEFINE DEBUG}
//------------------------------------------------------------------------------
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers;

function  PSSCheck(S, M, N: TInteger): Boolean;
procedure PSSKeys(N, P, Q, Up, Uq: TInteger; BSize: Longint);
procedure PSSSign(S, M, N, P, Q, Up, Uq: TInteger);


implementation
//==============================================================================
uses Windows, SysUtils, Integers_Primes, HIT_Hash, HIT_Rand;

////////////////////////////////////////////////////////////////////////////////
// CONSTANTS
////////////////////////////////////////////////////////////////////////////////

const
  //----- Message digest sizes
  PSS_MDSIZE_k0 = 160;
  PSS_MDSIZE_k1 = 128;
  //----- Hash functions G and H (Indexed RIPEMD-160)
  PSS_HFUNC_G = 0;
  PSS_HFUNC_H = 1; //Must be different from "G"

////////////////////////////////////////////////////////////////////////////////
// ROUTINES
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Return TRUE iff the P pointers are distinct
//------------------------------------------------------------------------------
function DistinctPointers(const P: array of Pointer): Boolean;
  var i, j : Longint;
begin
  Result := true;
  for i := 0 to High(P)-1 do
    for j := i+1 to High(P) do
      if P[i] = P[j] then
      begin
        Result := false;
        Exit;
      end;
end;

//==============================================================================
// Compute the public and secret keys for the Rabin PSS
// -> BSize, binary size of the modulus N
// <- P, prime and congruent to 3 modulo 4
// <- Q <> P, prime and congruent to 3 modulo 4
// <- N = P * Q
// <- Up, such that Up = 1 mod P and Up = 0 mod Q
// <- Uq, such that Uq = 0 mod P and Uq = 1 mod Q
//------------------------------------------------------------------------------
// Public key: N
// Secret key: N,P,Q,Up,Uq
//==============================================================================
procedure PSSKeys(N, P, Q, Up, Uq: TInteger; BSize: Longint);
  var
    rg     : TIntCSRG; //Secure Random Generator (See HIT_Rand.pas)
    sp, sq : Longint;
    OldCL  : Boolean;
begin
  //----- Check
  if not DistinctPointers([N,P,Q,Up,Uq]) then
    HITRaiseDuplicatePtr('PSSKeys','N,P,Q,Up,Uq');
  if (BSize < (PSS_MDSIZE_k0 + PSS_MDSIZE_k1)) or (BSize > 4096) then
    HITRaiseInvalidArg('PSSKeys',
                       '(BSize < ' + IntToStr(PSS_MDSIZE_k0 + PSS_MDSIZE_k1) +
                       ') or (BSize > 4096)');

  rg := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;
    //----- Create a secure random generator
    rg := TIntCSRG.Create;

    //----- Generate the prime P
    sp := Succ(BSize shr 1);
    rg.Rnd(P,sp);
    P.Bit[sp-2] := true; //So that N.BitSize = BSize
    IMakePrime(P,[2,3,5,7,11],3,4);

    //----- Generate the prime Q
    sq := BSize - P.BitSize;
    repeat
      rg.Rnd(Q,sq);
      Q.Bit[sq-2] := true;
      IMakePrime(Q,[2,3,5,7,11],3,4);
    until Q.BitSize = sq; //Always true on first iteration, but...

    //----- N := P * Q (N is a Blum integer)
    ISet(N,P);
    IMul(N,Q);

    //----- Chinese Remainder Theorem
    ISet(Uq,P);
    ISet(Up,Q);
    IInvMod(Uq,Up); //I = P^(-1) mod Q
    ISub(Up,Uq);    //-I mod Q (always positive)
    IMul(Up,P);
    IInc(Up);   //Up = (((0-1) * I) mod Q) * P + 1
    IMul(Uq,P); //Uq = (((1-0) * I) mod Q) * P + 0
  finally
    //----- Free the generator
    rg.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

//==============================================================================
// Rabin scheme: Signature of a message
// -> M, message to sign
// -> N,P,Q,Up,Uq, secret key of the signer
// <- S, signature of M
//==============================================================================
procedure PSSSign(S, M, N, P, Q, Up, Uq: TInteger);
  var
    rg               : TIntCSRG; //Secure Random Generator (See HIT_Rand.pas)
    B, C, R          : TInteger;
    k, k_k1, k_k0_k1 : Longint;
    OldCL            : Boolean;
begin
  //----- Check
  if not DistinctPointers([S,M,N,P,Q,Up,Uq]) then
    HITRaiseDuplicatePtr('PSSSign','S,M,N,P,Q,Up,Uq');
  if M.Neg then HITRaiseInvalidArg('PSSSign','M < 0');
  //----- k = maximal size of S
  k := N.BitSize - 1;
  if k <= (PSS_MDSIZE_k0 + PSS_MDSIZE_k1) then
    HITRaiseInvalidArg('PSSSign', 'N.BitSize <= ' +
                                        IntToStr(PSS_MDSIZE_k0+PSS_MDSIZE_k1));

  B := nil;
  C := nil;
  R := nil;
  rg := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;

    //----- Create the integers
    B := TInteger.Create;
    C := TInteger.Create;
    R := TInteger.Create;

    //----- Create a secure random generator
    rg := TIntCSRG.Create;

    k_k1 := k - PSS_MDSIZE_k1;
    k_k0_k1 := k_k1 - PSS_MDSIZE_k0;

    //===== Main loop
    repeat
      //----- R, random and secure, at most k0 bits
      rg.Rnd(R,PSS_MDSIZE_k0,false);

      /////////////////////////////////////////////////////////////////
      // !!! Here, it is possible to put a subliminal message into R //
      /////////////////////////////////////////////////////////////////

      //----- Concatenation, S := M * 2^k0 or R
      ISet(S,M);
      ICat(S,R,PSS_MDSIZE_k0);

      //----- S := H(M&R) (k1 bits) ("&" ~ concatenation operator)
      //      IH160(_,_,PSS_HFUNC_H) -> Indexed RIPEMD-160 hash function
      IH160(S,PSS_MDSIZE_k1,PSS_HFUNC_H);

      //----- B := G(H(M&R)) (k-k1 bits)
      ISet(B,S);
      IH160(B,k_k1,PSS_HFUNC_G);
      ISet(C,B);

      //----- B := (k0 leading bits of G(S)) xor R
      IShr(B,k_k0_k1);
      IXor(B,R);

      //----- C := k-k0-k1 trailing bits of G(S)
      ICut(C,k_k0_k1);

      //----- Concatenation, S := S&B&C (S has at most k bits)
      ICat(S,B,PSS_MDSIZE_k0);
      ICat(S,C,k_k0_k1);

      //----- Test whether S is a quadratic residue mod N
    until (IJacobi(S,P) = 1) and (IJacobi(S,Q) = 1);

    //----- Principal square roots mod P and Q
    ISet(R,S);
    ISqrRootMod(R,P); //Sp := S^((P+1)/4) mod P (Here, J(Sp,P) = 1)
    ISqrRootMod(S,Q); //Sq := S^((Q+1)/4) mod Q (Here, J(Sq,Q) = 1)

    //----- Random choice of one of the 4 square roots of S mod N
    rg.Rnd(B,2,false);
    if B.Bit[0] then ISubr(R,P); //Toss up for Jacobi(Sp,P) = 1
    if B.Bit[1] then ISubr(S,Q); //Toss up for Jacobi(Sq,Q) = 1

    //----- Square root of S mod N: S := (R*Up + S*Uq) mod N
    IMul(R,Up);
    IMul(S,Uq);
    IAdd(S,R);
    IMod(S,N);
  finally
    //----- Free the random generator
    rg.Free;
    //----- Free the numbers
    B.Free;
    C.Free;
    R.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

//==============================================================================
// Rabin scheme: Checking of the signature
// -> S, signature of M
// -> M, signed message
// -> N, public key of the signer
// <- TRUE if and only if S is a valid signature
//==============================================================================
function PSSCheck(S, M, N: TInteger): Boolean;
  var
    A, B, C, X, Y    : TInteger;
    k, k_k1, k_k0_k1 : Longint;
    OldCL            : Boolean;
begin
  //----- Check
  if not DistinctPointers([S,M,N]) then
    HITRaiseDuplicatePtr('PSSCheck','S,M,N');
  //----- k = maximal size of S
  k := N.BitSize - 1;
  if k <= (PSS_MDSIZE_k0 + PSS_MDSIZE_k1) then
    HITRaiseInvalidArg('PSSCheck', 'N.BitSize <= ' +
                                        IntToStr(PSS_MDSIZE_k0+PSS_MDSIZE_k1));

  Result := false;
  A := nil;
  B := nil;
  C := nil;
  X := nil;
  Y := nil;
  OldCL := HITCryptoLevel;
  try
    //----- Secure memory management
    HITCryptoLevel := true;

    //----- A := S mod N
    A := TInteger.Create(S);
    ISqr(A);
    IMod(A,N);
    B := TInteger.Create(A);
    C := TInteger.Create(A);

    k_k1 := k - PSS_MDSIZE_k1;
    k_k0_k1 := k_k1 - PSS_MDSIZE_k0;

    //----- Split: S = A&B&C ("&" ~ concatenation operator)
    IShr(A,k_k1);    //A <- k1 leading bits of S
    ICut(C,k_k0_k1); //C <- k-k0-k1 trailing bits of S

    //----- X := G1(A) (X has at most k-k1 bits)
    X := TInteger.Create(A);
    IH160(X,k_k1,PSS_HFUNC_G);
    Y := TInteger.Create(X);
    ICut(X,k_k0_k1);
    if not IEqu(X,C) then Exit; //Invalid signature

    //----- B <- (k0 intermediate bits of S) xor (Y shr (k-k0-k1))
    IShr(B,k_k0_k1);
    ICut(B,PSS_MDSIZE_k0);
    IShr(Y,k_k0_k1);
    IXor(B,Y);

    //----- Concatenation, Y := M&B
    ISet(Y,M);
    ICat(Y,B,PSS_MDSIZE_k0);

    //----- Y := H(M&B) (k1 bits)
    IH160(Y,PSS_MDSIZE_k1,PSS_HFUNC_H);
    Result := IEqu(Y,A);
  finally
    //----- Free the numbers
    A.Free;
    B.Free;
    C.Free;
    X.Free;
    Y.Free;
    //----- Restore initial value
    HITCryptoLevel := OldCL;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

